# Database Lib
"""
Oracle
PostGresSQL
SQLite
SQLServer
Hive
Spark
"""
import os, datetime, pandas, time, re
from collections import namedtuple, OrderedDict

import jmespath
import sqlalchemy
from multiprocessing import Queue, Process

from xutil.helpers import (
  log,
  elog,
  slog,
  get_exception_message,
  struct,
  now,
  get_databases,
  get_dir_path,
  get_profile,
  get_variables,
  file_exists,
  str_rmv_indent,
  ptable,
  make_rec,
  get_error_str,
)
from xutil.diskio import read_yaml, write_csvs

conns = {}

_fwklike = lambda k, v: "lower({}) like lower('{}')".format(k, v)
_fwkeq = lambda k, v: "{} = '{}'".format(k, v)
_fw = lambda sep, _fwkop, **kws: sep.join([_fwkop(k, v) for k, v in kws.items()])  # Format WHERE
fwa = lambda _fwkop=_fwkeq, **kws: _fw(' and ', _fwkop, **kws)  # Format WHERE AND
fwo = lambda _fwkop=_fwkeq, **kws: _fw(' or ', _fwkop, **kws)  # Format WHERE OR
rows_to_dicts = lambda rows: [row._asdict() for row in rows]


class DBConn(object):
  """Base class for database connections"""

  _fix_f_name = lambda self, f: f
  _to_text = lambda self, t: t

  def __init__(self, conn_dict, profile=None, echo=False):
    "Inititate connection"
    self._cred = struct(conn_dict)
    self._cred.kwargs = conn_dict.get('kwargs', {})
    self.name = self._cred.get('name', None)
    self.username = self._cred.get('username', None)
    self.type = self._cred.type
    self.engine = None
    self._cursor_description = None
    self.profile = profile
    self.batch_size = 10000
    self.fetch_size = 20000
    self.echo = echo
    self.connect()
    self.last_connect = now()

    # Base Template
    template_base_path = '{}/database/templates/base.yaml'.format(
      get_dir_path())
    self.template_dict = read_yaml(template_base_path)

    # Specific Type Template
    template_path = '{}/database/templates/{}.yaml'.format(
      get_dir_path(), self.type)
    temp_dict = read_yaml(template_path)

    for key1 in temp_dict:
      # Level 1
      if isinstance(temp_dict[key1], dict):
        if key1 not in self.template_dict:
          self.template_dict[key1] = temp_dict[key1]

        # Level 2
        for key2 in temp_dict[key1]:
          # Always Overwrite
          self.template_dict[key1][key2] = temp_dict[key1][key2]
      else:
        # Level 1 Non-Dict Overwrite
        self.template_dict[key1] = temp_dict[key1]

    self.variables = self._template('variables')

    if os.getenv('PROFILE_YAML'):
      other_vars = get_variables()
      for key in other_vars:
        self.variables[key] = other_vars[key]

    self.tmp_folder = self.variables['tmp_folder']
    self.set_variables()

    if echo:
      log("Connected to {} as {}".format(self._cred.name, self._cred.user))

  def connect(self):
    """Connect to Database"""
    self.engine = self.get_engine()
    self.connection = self.engine.connect()
  
  def close(self):
    """Close database connection"""
    self.conn.connection.close()

  def reconnect(self, min_tresh=0):
    """Re-Connect to Database if minute threshold reached"""
    if (now() - self.last_connect).total_seconds() > min_tresh * 60:
      log('Reconnecting to {}...'.format(self.name))
      self.connect()
      self.last_connect = now()

  def set_variables(self):
    """Set custom variables"""
    raise Exception("Method 'set_variables' is not implemented!")

  def get_dialect(self, echo=False):
    """SQLAlchemy dialect"""
    raise Exception("Method 'get_dialect' is not implemented!")

  def get_engine(self, echo=False):
    import sqlalchemy
    if not self.engine:
      self.create_engine(echo=self.echo)
    self.engine_inspect = sqlalchemy.inspect(self.engine)
    return self.engine

  def check_pk(self, table, fields):
    "Check Primary key to ensure there are not duplicates"
    if 'where' in fields.lower():
      fields, where_clause = fields.lower().split('where')
      where_clause = 'where ' + where_clause
    else:
      where_clause = ''
    sql = '''
    select
      '{table}' as table,
      case when count(1) = count({fields}) then 'PASS' else 'FAIL' end as pk_result
    from {table}
    {where_clause}
    '''.format(
      table=table,
      fields=fields,
      where_clause=where_clause,
    )
    data = self.query(sql, echo=False)
    headers = self._fields
    print(ptable(headers, data))
    if data[0].pk_result == 'FAIL':
      raise (Exception('PK Text failed for table "{}" with fields "{}"'.format(
        table, fields)))

  def _do_execute(self, sql):
    try:
      self._cursor_description = None
      self.fields = None
      self.result = self.connection.execute(sql)
      self._cursor_description = self.result._cursor_description()
      self._fields = self._get_cursor_fields()
    except Exception as E:
      if 'not open' in get_error_str(E):
        pass # error when Oracle doesn't have a cursor open
      else:
        log(Exception('Error for SQL:\n' + sql))
        raise E

  def execute_multi(self,
                    sql,
                    dtype='namedtuple',
                    limit=None,
                    echo=True,
                    query_name='Record',
                    log=log):
    """
    Execute multiple SQL statements separtated by ';'. Returns a generator.
    Example:
      for fields, rows in conn.execute(sql):
        print(fields)
        print(len(rows))
    """

    self.reconnect(min_tresh=10)

    data = None
    fields = None
    rows = []
    message_mapping = {
      'drop ': 'Dropping {}.',
      'truncate ': 'Truncating {}.',
      'select ': 'Selecting {}.',
      'create ': 'Creating {}.',
      'insert ': 'Inserting {}.',
      'alter ': 'Altering {}.',
      'update ': 'Updating {}.',
      'delete ': 'Deleting {}.',
      'exec ': 'Calling Procedure {}.',
      'grant ': 'Granting {}.',
    }

    sqls = sql.split(';')

    for sql in sqls:
      if not sql.strip(): continue

      sql_ = sql.strip().lower()

      for word, message in message_mapping.items():
        if sql_.startswith(word):
          if echo:
            log(
              message.format(' '.join(
                sql_.splitlines()[0].split()[1:3]).upper()))
          break

      # Call procedure with callproc
      if sql_.startswith('exec '):
        procedure = sql_[5:].split('(')[0]
        args = sql_[5:].split('(')[1][:-1].replace("'", '').split(',')
        args = [a.strip() for a in args]
        cursor.callproc(procedure, args)
        continue

      try:
        self._fields = []
        rows = self.query(
          sql,
          rec_name=query_name,
          dtype=dtype,
          limit=limit,
          echo=echo,
          log=log)
        fields = self._fields

        if '-- pk_test:' in sql.lower() and sql_.startswith('create'):
          sql_lines = sql_.splitlines()
          regexp = r'create\s+table\s+(\S*)[\sa-zA-Z\d]+ as'
          table = re.findall(regexp, sql_lines[0])[0]
          line = [
            l for l in sql_lines if l.strip().lower().startswith('-- pk_test:')
          ][0]
          fields = line.split(':')[-1]
          self.check_pk(table, fields)

      except Exception as E:
        message = get_exception_message().lower()

        if sql_.startswith(
            'drop ') and self.error_msg['table_not_exist'] in message:
          log("WARNING: Table already dropped.")
        else:
          raise E

      if not fields: fields = []

      yield fields, rows

  def execute(self,
              sql,
              dtype='tuple',
              limit=None,
              echo=True,
              query_name='Record',
              log=log):
    """Execute SQL, return last result"""
    self.reconnect(min_tresh=10)

    data = None
    fields = None
    rows = []
    message_mapping = {
      'drop ': 'Dropping {}.',
      'truncate ': 'Truncating {}.',
      'select ': 'Selecting {}.',
      'create ': 'Creating {}.',
      'insert ': 'Inserting {}.',
      'alter ': 'Altering {}.',
      'update ': 'Updating {}.',
      'delete ': 'Deleting {}.',
      'exec ': 'Calling Procedure {}.',
      'grant ': 'Granting {}.',
    }

    sql_ = sql.strip().lower()

    for word, message in message_mapping.items():
      if sql_.startswith(word):
        if echo:
          log(
            message.format(' '.join(
              sql_.splitlines()[0].split()[1:3]).upper()))
        break

    # Call procedure with callproc
    if sql_.startswith('exec '):
      procedure = sql_[5:].split('(')[0]
      args = sql_[5:].split('(')[1][:-1].replace("'", '').split(',')
      args = [a.strip() for a in args]
      connection = self.engine.raw_connection()
      try:
        cursor = connection.cursor()
        cursor.callproc(procedure, args)
        self._fields = self._get_cursor_fields(cursor_desc=cursor.description)
        rows = list(cursor.fetchall())
        cursor.close()
        connection.commit()
        return fields, rows
      finally:
        connection.close()

    try:
      self._fields = []
      rows = self.query(
        sql,
        rec_name=query_name,
        dtype=dtype,
        limit=limit,
        echo=echo,
        log=log)
      fields = self._fields

      if '-- pk_test:' in sql.lower() and sql_.startswith('create'):
        sql_lines = sql_.splitlines()
        regexp = r'create\s+table\s+(\S*)[\sa-zA-Z\d]+ as'
        table = re.findall(regexp, sql_lines[0])[0]
        line = [
          l for l in sql_lines if l.strip().lower().startswith('-- pk_test:')
        ][0]
        fields = line.split(':')[-1]
        self.check_pk(table, fields)

    except Exception as E:
      message = get_exception_message().lower()

      if sql_.startswith(
          'drop ') and self.error_msg['table_not_exist'] in message:
        log("WARNING: Table already dropped.")
      else:
        raise E

    if not fields: fields = []

    return fields, rows

  def insert(self, table, data, echo=False):
    """Insert records of namedtuple or dicts"""
    raise Exception('insert not implemented')

  def drop_table(self, table, log=log):
    "Drop table"

    try:
      sql = self._template('core.drop_table').format(table)
      self._do_execute(sql)
    except Exception as E:
      message = get_exception_message().lower()
      if self._template('error_filter.table_not_exist') in message:
        if self.echo:
          log('Table "{}" already dropped.'.format(table))
      else:
        raise E

  def create_table(self, table, field_types, drop=False, log=log):
    "Create table"

    if drop:
      self.drop_table(table, log=log)

    new_ftypes = OrderedDict()
    for f in field_types:
      ftype, max_len, dec_len = field_types[f]
      if dec_len:
        suff = '({},{})'.format(max_len, dec_len)
      elif max_len:
        suff = '({})'.format(max_len)
      else:
        suff = ''

      new_ftypes[f] = self._template('general_type_map')[ftype].replace(
        '()', suff)

    field_types_str = ', \n'.join([
      self._fix_f_name(field) + ' ' + new_ftypes[field] for field in new_ftypes
    ])

    sql = self._template('core.create_table').format(
      table=table,
      col_types=field_types_str,
    )

    # log('Creating table: \n' + sql))
    try:
      self._do_execute(sql)
    except Exception as e:
      raise e

    log('Created table "{}"'.format(table))

  def _get_cursor_fields(self, as_dict=False, native_type=True, cursor_desc=None):
    "Get fields of active Select cursor"
    fields = OrderedDict()
    cursor_desc = cursor_desc if cursor_desc else self._cursor_description
    if cursor_desc == None:
      return []

    for f in cursor_desc:
      f_name = f[0].lower()
      if as_dict:
        if native_type:
          f_type = f[1]
        else:
          f_type = self.reverse_data_map[f[1]]

          # assign floa/double as needed
          if 'cx_Oracle.NUMBER' in str(f[1]):
            if f[4] and f[4] > 11: f_type = 'long'
            if f[5] and f[5] > 0: f_type = 'double'

        fields[f_name] = f_type
      else:
        fields[f_name] = None

    if as_dict:
      return fields
    else:
      return list(fields.keys())

  def stream(self,
             sql,
             rec_name='Record',
             dtype='namedtuple',
             yield_chuncks=False,
             chunk_size=None,
             limit=None,
             echo=True):
    "Stream Select from SQL, yield records as they come in"
    self.reconnect(min_tresh=10)
    if echo: log("Streaming SQL for '{}'.".format(rec_name))

    fetch_size = limit if limit else self.fetch_size
    fetch_size = chunk_size if chunk_size else fetch_size

    try:
      self._do_execute(sql)
    except Exception as e:
      raise e

    if dtype == 'tuple':
      make_rec = lambda row: row
      make_batch = lambda rows: rows
    elif dtype == 'dataframe':
      yield_chuncks=True
      make_batch = lambda rows: pandas.DataFrame(rows, columns=self._fields)
    else:
      Record = namedtuple(
        rec_name.replace(' ', '_').replace('.', '_'), self._fields)
      make_rec = lambda row: Record(*row)
      make_batch = lambda rows: [make_rec(r) for r in rows]

    self._stream_counter = 0

    while True:
      if not self._fields:
        break
      rows = self.result.fetchmany(fetch_size)
      if rows:
        if yield_chuncks:
          batch = make_batch(rows)
          self._stream_counter += len(batch)
          if len(batch):
            yield batch
        else:
          for row in rows:
            self._stream_counter += 1
            yield make_rec(row)
      else:
        break
      if limit:
        break

    # log('Stream finished at {} records.'.format(self._stream_counter))

  def query(self,
             sql,
             rec_name='Record',
             dtype='namedtuple',
             limit=None,
             echo=True,
             retrying=False,
             log=log):
    "Select from SQL, return list of namedtuples"
    # if echo: log("Running SQL for '{}'.".format(rec_name))

    self.reconnect(min_tresh=10)
    s_t = datetime.datetime.now()

    def get_rows(cursor):
      counter = 0
      row = cursor.fetchone()
      while row:
        counter += 1
        yield row
        if limit and counter == limit:
          break
        row = cursor.fetchone()

    _data = list(self.stream(sql, dtype=dtype, echo=False, limit=limit))

    fields = self._fields
    if not fields: return []

    if dtype == 'namedtuple':
      Record = namedtuple(rec_name.replace(' ', '_').replace('.', '_'), fields)
      if limit:
        data = [Record(*row) for row in _data]
      else:
        data = [Record(*row) for row in _data]

    elif dtype == 'tuple':
      if limit:
        data = [tuple(row) for row in _data]
      else:
        data = [tuple(row) for row in _data]

    elif dtype == 'dataframe':
      if limit:
        data = pandas.DataFrame([row for row in _data], columns=fields)
      else:
        data = pandas.DataFrame([row for row in _data], columns=fields)
    else:
      raise (Exception('{} is not recongnized.'.format(dtype)))

    secs = (datetime.datetime.now() - s_t).total_seconds()
    rate = round(len(data) / secs, 1)
    if echo:
      log(" >>> Got {} rows in {} secs [{} r/s].".format(
        len(data), secs, rate))
    return data

  def _split_schema_table(self, table_name):
    schema, table = table_name.split('.') if '.' in table_name else (
      self.username, table_name)
    return schema, table

  def _concat_fields(self, fields, as_text=False):
    return ' || '.join(fields)

  def _template(self, template_key_str):
    val = jmespath.search(template_key_str, self.template_dict)
    if isinstance(val, str):
      val = str_rmv_indent(val)
    return val

  def get_schemas(self, echo=True):
    "Get list of schemas."
    Rec = namedtuple('Schemas', 'schema')
    self._fields = Rec._fields

    sql_tmpl = self._template('metadata.schemas')
    if sql_tmpl:
      schemas = [r[0] for r in self.query(sql_tmpl)]
    else:
      # http://docs.sqlalchemy.org/en/rel_0_9/core/reflection.html#sqlalchemy.engine.reflection.Inspector.get_schemas
      self.get_engine(echo=echo)
      schemas = self.engine_inspect.get_schema_names()

    rows = [Rec(s) for s in schemas]
    return rows

  def get_objects(self, schema, object_type='all', echo=True):
    "Get metadata for objects. object_type in 'all', 'table', 'view'"
    Rec = namedtuple('Table', 'schema object_name object_type')
    self._fields = Rec._fields

    def get_rec(object_name, object_type):
      r_dict = dict(
        schema=schema, object_name=object_name, object_type=object_type)
      return Rec(**r_dict)

    if object_type == 'all':
      table_rows = self.get_tables(schema)
      rows = [get_rec(r.table, 'table') for r in sorted(table_rows)]
      view_rows = self.get_views(schema)
      rows += [get_rec(r.view, 'view') for r in sorted(view_rows)]
    elif object_type == 'table':
      table_rows = self.get_tables(schema)
      rows = [get_rec(r.table, 'table') for r in sorted(table_rows)]
    elif object_type == 'view':
      view_rows = self.get_views(schema)
      rows += [get_rec(r.view, 'view') for r in sorted(view_rows)]
    else:
      raise Exception('Object type "{}" not supported!'.format(object_type))

    return rows

  def get_tables(self, schema, echo=True):
    "Get metadata for tables."
    Rec = namedtuple('Table', 'schema table')
    self._fields = Rec._fields

    def get_rec(table):
      r_dict = dict(schema=schema, table=table)
      return Rec(**r_dict)

    sql_tmpl = self._template('metadata.tables')
    if sql_tmpl:
      tables = self.query(sql_tmpl.format(schema=schema))
      if hasattr(self, '_std_get_tables'):
        tables = self._std_get_tables(schema, tables)
    else:
      self.get_engine(echo=echo)
      tables = self.engine_inspect.get_table_names(schema)

    rows = [get_rec(t) for t in sorted(tables)]
    self._fields = Rec._fields
    return rows

  def get_views(self, schema, echo=True):
    "Get metadata for views."
    Rec = namedtuple('View', 'schema view')
    self._fields = Rec._fields

    def get_rec(view):
      r_dict = dict(schema=schema, view=view)
      return Rec(**r_dict)

    sql_tmpl = self._template('metadata.views')
    if sql_tmpl:
      views = [r[0] for r in self.query(sql_tmpl.format(schema=schema))]
    else:
      self.get_engine(echo=echo)
      views = self.engine_inspect.get_view_names(schema)

    rows = [get_rec(v) for v in sorted(views)]
    self._fields = Rec._fields
    return rows

  def get_columns(self,
                  table_name,
                  object_type=None,
                  echo=False,
                  include_schema_table=True,
                  native_type=True):
    "Get column metadata for table"
    if include_schema_table:
      headers = 'schema table id column_name type nullable default autoincrement'
    else:
      headers = 'id column_name type  nullable default autoincrement'

    Rec = namedtuple('Columns', headers)
    self._fields = Rec._fields
    all_rows = []

    table_names = table_name if isinstance(table_name, list) else [table_name]

    for table_name in table_names:
      schema, table = self._split_schema_table(table_name)

      def get_rec(r_dict, column_order):
        if include_schema_table:
          r_dict['schema'] = schema
          r_dict['table'] = table
        r_dict['column_name'] = r_dict['name']
        r_dict['type'] = str(r_dict['type'])
        if not native_type:
          r_dict['type']= r_dict['type'].lower()
          r_dict['type'] = r_dict['type'].split('(')[0] if '(' in r_dict[
            'type'] else r_dict['type']
          native_type_map = self._template('native_type_map')
          if not r_dict['type'] in native_type_map:
            raise Exception('Field type "{}" not in native_type_map for {}'.format(r_dict['type'], self.type))
          r_dict['type'] = native_type_map[r_dict['type']]
        r_dict['id'] = column_order

        for k in list(r_dict):
          if k not in headers.split():
            del r_dict[k]

        if '(' in r_dict['type']:
          r_dict['type'] = r_dict['type'].split('(')[0]

        return Rec(**r_dict)

      sql_tmpl = self._template('metadata.columns')
      if sql_tmpl:
        rows = self.query(sql_tmpl.format(table=table, schema=schema))
        if hasattr(self, '_std_get_columns'):
          rows = self._std_get_columns(schema, table, rows)
      else:
        self.get_engine(echo=echo)
        rows = self.engine_inspect.get_columns(table, schema=schema)

      all_rows += [get_rec(r_dict, i + 1) for i, r_dict in enumerate(rows)]

    self._fields = Rec._fields
    return all_rows

  def get_primary_keys(self, table_name, echo=False):
    "Get PK metadata for table"
    Rec = namedtuple('PKs', 'schema table pk_name column_name column_order')
    self._fields = Rec._fields
    schema, table = self._split_schema_table(table_name)

    def get_rec(col, pk_name, column_order):
      r_dict = {}
      r_dict['schema'] = schema
      r_dict['table'] = table
      r_dict['pk_name'] = pk_name
      r_dict['column_name'] = col
      r_dict['column_order'] = column_order
      return Rec(**r_dict)

    sql_tmpl = self._template('metadata.primary_keys')
    if sql_tmpl:
      rows = self.query(sql_tmpl.format(table=table, schema=schema))
    else:
      self.get_engine(echo=echo)
      r_dict = self.engine_inspect.get_pk_constraint(table, schema=schema)
      rows = [
        get_rec(col, r_dict['name'], i + 1)
        for i, col in enumerate(r_dict['constrained_columns'])
      ]

    return rows

  def get_indexes(self, table_name, echo=False):
    "Get indexes metadata for table"
    Rec = namedtuple(
      'Indexes', 'schema table index_name column_name column_order unique')
    self._fields = Rec._fields
    schema, table = self._split_schema_table(table_name)

    def get_rec(r_dict):
      r_dict['schema'] = schema
      r_dict['table'] = table
      r_dict['index_name'] = r_dict['name']
      r_dict['unique'] = str(r_dict['unique'])
      del r_dict['name']
      for i, col in enumerate(r_dict['column_names']):
        r_dict['column_name'] = col
        r_dict['column_order'] = i + 1
        yield Rec(**r_dict)

    sql_tmpl = self._template('metadata.indexes')
    if sql_tmpl:
      rows = self.query(sql_tmpl.format(table=table, schema=schema))
    else:
      self.get_engine(echo=echo)
      rows = self.engine_inspect.get_indexes(table, schema=schema)
      rows = [get_rec(r_dict) for r_dict in rows]

    return rows

  def get_ddl(self, table_name, object_type=None, echo=True):
    "Get ddl for table"
    Rec = namedtuple('DDL', 'ddl')
    self._fields = Rec._fields
    schema, table = self._split_schema_table(table_name)

    sql_tmpl = self._template('metadata.ddl')
    if sql_tmpl:
      rows = self.query(
        sql_tmpl.format(
          schema=schema,
          table=table,
          obj_type=object_type,
        ))
    else:
      self.get_engine(echo=echo)
      ddl = self.engine_inspect.get_view_definition(table, schema=schema)
      rows = [Rec(ddl)] if ddl else []

    self._fields = Rec._fields
    return rows

  def get_all_columns(self):
    "Get all columns for all tables / views"
    sql_tmpl = self._template('metadata.all_columns')
    if not sql_tmpl:
      raise Exception('get_all_columns not implemented for {}'.format(
        self.type))

    rows = self.query(sql_tmpl)
    return rows

  def get_all_tables(self):
    "Get all tables / views"
    sql_tmpl = self._template('metadata.all_tables')
    if not sql_tmpl:
      raise Exception('get_all_tables not implemented for {}'.format(
        self.type))

    rows = self.query(sql_tmpl)
    return rows

  def analyze_fields(self,
                     analysis,
                     table_name,
                     fields=[],
                     as_sql=False,
                     union=True,
                     expr_func_map={},
                     **kwargs):
    """Base function for field level analysis
      expr_func_map: contains mapping for expression to SQL function to all fields
    """
    if '.' not in table_name:
      raise Exception("table_name must have schema and name in it with a '.'")
    if analysis not in self.template_dict['analysis']:
      raise Exception("'{}' not found in template for '{}'.".format(
        analysis, self.type))

    schema, table = self._split_schema_table(table_name)

    # get field type
    field_rows = self.get_columns(table_name)
    field_type = {r.column_name.lower(): r.type for r in field_rows}

    if not fields:
      fields = [r.column_name for r in field_rows]

    for expr in list(expr_func_map):
      tmpl_path = 'function.' + expr_func_map[expr]
      expr_func_map[expr] = ',\n'.join([
        self._template(tmpl_path).format(field=field)
        for field in [r.column_name for r in field_rows]
      ])

    sep = ' \nunion all\n' if union else ' \n ;\n'
    sql = sep.join([
      self._template('analysis.' + analysis).format(
        schema=schema,
        field=field,
        table=table,
        type=field_type[field.lower()] if field else '',
        **expr_func_map,
        **kwargs) for field in fields
    ])
    return sql if as_sql else self.query(sql, analysis, echo=False)

  def analyze_tables(self, analysis, tables=[], as_sql=False, **kwargs):
    """Base function for table level analysis"""
    if analysis not in self.template_dict['analysis']:
      raise Exception("'{}' not found in template for '{}'.".format(
        analysis, self.type))

    if not tables and 'schema' in kwargs:
      # get all tables
      rows = self.get_schemas(kwargs['schema'])
      crt_obj = lambda r: struct(dict(schema=r.schema, table=r.object_name))
      objs = [crt_obj(r) for r in rows]
    else:
      crt_obj = lambda schema, table: struct(dict(schema=schema, table=table))
      objs = [crt_obj(*self._split_schema_table(t)) for t in tables]

    sql = ' \nunion all\n'.join([
      self._template('analysis.' + analysis).format(
        schema=obj.schema, table=obj.table, **kwargs) for obj in objs
    ])

    return sql if as_sql else self.query(sql, analysis, echo=False)


  def analyze_join_match(self,
                        t1,
                        t2,
                        t1_field,
                        t2_field,
                        t1_filter='1=1',
                        t2_filter='1=1',
                        as_sql=False,
                        as_text=True,
                        lowercase=True):
    def get_kwargs(t1, t2, t1_field, t2_field, t1_filter, t2_filter):
      t1_field_arr = ['t1.' + f for f in t1_field.split(',')]
      t2_field_arr = ['t2.' + f for f in t2_field.split(',')]
      t1_field_concat = self._concat_fields(t1_field_arr, as_text=as_text)
      t2_field_concat = self._concat_fields(t2_field_arr, as_text=as_text)
      to_text = self._to_text

      if lowercase:
        conds = ' and '.join([
          'lower({}) = lower({})'.format(to_text(f), to_text(t2_field_arr[i]))
          for i, f in enumerate(t1_field_arr)
        ])
      else:
        conds = ' and '.join([
          '{} = {}'.format(to_text(f), to_text(t2_field_arr[i]))
          for i, f in enumerate(t1_field_arr)
        ])
      t1_fields1 = t1_field
      t2_fields1 = t2_field
      t1_field = ', '.join(['t1.' + f for f in t1_field_arr])
      t2_field = ', '.join(['t2.' + f for f in t2_field_arr])

      return dict(
        t1=t1,
        t1_field=t1_field_concat,
        t1_fields1=t1_fields1,
        t1_filter=t1_filter,
        t2=t2,
        t2_field=t2_field_concat,
        t2_fields1=t2_fields1,
        t2_filter=t2_filter,
        conds=conds,
      )

    kwargs = get_kwargs(
      t1=t1,
      t2=t2,
      t1_field=t1_field,
      t2_field=t2_field,
      t1_filter=t1_filter,
      t2_filter=t2_filter,
    )
    sql = self.analyze_fields(
      'table_join_match', t1, [''], as_sql=True, **kwargs)

    return sql if as_sql else self.query(sql, 'table_join_match', echo=False)


def get_conn(db,
             dbs=None,
             echo=True,
             reconnect=False,
             use_jdbc=False,
             conn_expire_min=10,
             spark_hive=False) -> DBConn:
  global conns

  dbs = dbs if dbs else get_databases()
  profile = get_profile()
  db_dict = struct(dbs[db])

  if db_dict.type.lower() == 'hive' and spark_hive:
    db_dict.type = 'spark'

  use_jdbc = True if (use_jdbc or ('use_jdbc' in db_dict
                                   and db_dict['use_jdbc'])) else use_jdbc

  if db in conns and not reconnect:
    if (now() - conns[db].last_connect).total_seconds() / 60 < conn_expire_min:
      return conns[db]

  if use_jdbc:
    log('*USING JDBC for ' + db)
    from .jdbc import JdbcConn
    conn = JdbcConn(db_dict, profile=profile)

  elif db_dict.type.lower() == 'oracle':
    from .oracle import OracleConn
    conn = OracleConn(db_dict, echo=echo)

  elif db_dict.type.lower() == 'spark':
    from .spark import SparkConn
    conn = SparkConn(db_dict, echo=echo)

  elif db_dict.type.lower() == 'hive':
    from .hive import HiveConn, Beeline
    if 'use_beeline' in db_dict and db_dict.use_beeline:
      conn = Beeline(db_dict, echo=echo)
    else:
      conn = HiveConn(db_dict, echo=echo)

  elif db_dict.type.lower() in ('postgresql', 'redshift'):
    from .postgresql import PostgreSQLConn
    conn = PostgreSQLConn(db_dict, echo=echo)

  elif db_dict.type.lower() == 'sqlserver':
    from .sqlserver import SQLServerConn
    conn = SQLServerConn(db_dict, echo=echo)

  elif db_dict.type.lower() == 'sqlite':
    from .sqlite import SQLiteConn
    conn = SQLiteConn(db_dict, echo=echo)
  else:
    raise Exception(f'Type {db_dict.type} not handled!')

  conns[db] = conn
  return conn


class SqlX:
  """
  SQL Express functions. Supports CRUD transactional operations.

  Suppose there is a table named 'cache', sqlx allows:
  
  sqlx.x('cache').insert(rows)
  sqlx.x('cache').insert_one(row)
  sqlx.x('cache').add(**kws)
  sqlx.x('cache').delete(where)
  sqlx.x('cache').update(rows, pk_fields)
  sqlx.x('cache').update_one(row, pk_cols)
  sqlx.x('cache').replace(rows, pk_fields)
  sqlx.x('cache').query(where)
  sqlx.x('cache').select_one(where)
  """

  def __init__(self, conn: DBConn, table, schema, ntRec: namedtuple):
    self.conn = conn
    self.table = table
    self.schema = schema
    self.ntRec = ntRec
    self.pk_fields = None
    self.table_obj = schema + '.' + table if schema else table

    self.insert_one = lambda row: self.insert([row])
    self.add = lambda **kws: self.insert([self.ntRec(**kws)])
    self.update_one = lambda row, pk_cols=None: self.update([row], pk_cols)
    self.update_rec=lambda pk_cols=None, **kws: self.update([make_rec(**kws)], pk_cols)
    self.replace_one = lambda row, pk_cols=None: self.replace([row], pk_cols)
    self.replace_rec=lambda pk_cols=None, **kws: self.replace([make_rec(**kws)], pk_cols)
    # self.select_one = lambda where: self.select_one(where, one=True)

  def _get_pk(self):
    if not self.pk_fields:
      pk_rows = self.conn.get_primary_keys(self.table_obj)
      self.pk_fields = [r.column_name for r in pk_rows]
    return self.pk_fields

  def insert(self, data):
    return self.conn.insert(self.table_obj, data)

  def update(self, data, pk_fields=None):
    if not pk_fields:
      pk_fields = self._get_pk()
      if not pk_fields:
        raise Exception("Need Keys to perform UPDATE!")
      t_fields = [x.lower() for x in data[0]._fields]
      for f in pk_fields:
        if not f.lower() in t_fields:
          # if keys not provided, need to make sure PK values are provided in data records
          raise Exception(
            "Value of  PK field '{}' must be provided to perform UPDATE!".
            format(f))

    self.conn.update(self.table_obj, data, pk_fields, echo=False)

  def update_one(self, row, pk_cols=None):
    self.update([row], pk_cols)

  def update_rec(self, pk_cols=None, **kws):
    self.update([make_rec(**kws)], pk_cols)

  def replace(self, data, pk_fields=None):
    if not pk_fields:
      pk_fields = self._get_pk()
    self.conn.replace(self.table_obj, data, pk_fields, echo=False)

  # def replace_rec(self, pk_cols=None, **kws):
  #   # add default None?
  #   for field in self.ntRec._fields:
  #     kws[field] = kws.get(field, None)

  #   self.replace([self.ntRec(**kws)], pk_cols)

  def query(self, where='1=1', one=False, limit=None, as_dict=False):
    rows = self.conn.query(
      "select * from {} where {}".format(self.table_obj, where),
      echo=False,
      limit=limit)
    rows = rows_to_dicts(rows) if as_dict else rows
    if one: return rows[0] if rows else None
    else: return rows

  def select_one(self, where, field=None, as_dict=False):
    row = self.query(where, one=True, as_dict=as_dict)
    if field and row:
      return row[field] if as_dict else row.__getattribute__(field)
    return row

  def delete(self, where):
    self.conn.execute("delete from {} where {}".format(self.table_obj, where))


def make_sqlx(conn, schema, tables):
  "Make sqlx lookup function for given tables"

  table_func_map = {}

  for table in tables:
    ntRec = namedtuple(table, tables[table].columns.keys())
    table_func_map[table] = SqlX(conn, table, schema, ntRec)

  # return table_func_map

  def sqlx(expr) -> SqlX:
    obj = jmespath.search(expr, table_func_map)
    if not obj:
      raise Exception('sqlx: Cannot find "{}"'.format(expr))
    return obj

  return sqlx


def get_sql_sources(sql_text, echo=False):
  """Obtain the source tables of a query
  """

  import sqlparse

  # replace "as(" to "as (" # this trips up the sql parser in CTEs
  sql_text = re.sub(r"as\(", "as (", sql_text, 0, re.MULTILINE | re.IGNORECASE)

  statements = sqlparse.parse(sql_text)
  cte_aliases = set()
  sql_sources = {}

  def get_sources(statement):
    sources_dict = {}
    last_kw_from = False
    last_kw_join = False
    cte_mode = False
    last_tok = None
    done = False

    while not done:
      for tok in statement.tokens:

        if tok.is_group:
          if cte_mode and isinstance(tok, sqlparse.sql.IdentifierList):
            for tok2 in tok.tokens:
              if isinstance(tok2, sqlparse.sql.Identifier):
                for tok3 in tok2.tokens:
                  if isinstance(tok3, sqlparse.sql.Parenthesis):
                    cte_aliases.add(tok3.parent.normalized.lower())
                    sources_dict2 = get_sources(tok3)
                    sources_dict = {**sources_dict, **sources_dict2}
          elif isinstance(tok, sqlparse.sql.Parenthesis):
            sources_dict2 = get_sources(tok)
            sources_dict = {**sources_dict, **sources_dict2}
          else:
            for tok2 in tok.tokens:
              if isinstance(tok2, sqlparse.sql.Parenthesis):
                cte_aliases.add(tok2.parent.normalized.lower())
                sources_dict2 = get_sources(tok2)
                sources_dict = {**sources_dict, **sources_dict2}


        if (last_kw_from or last_kw_join) and last_tok.is_whitespace:
          if isinstance(tok, sqlparse.sql.IdentifierList):
            for tok2 in tok.tokens:
              if isinstance(tok2, sqlparse.sql.Identifier) and '(' in tok2.value:
                sources_dict2 = get_sources(tok2)
                sources_dict = {**sources_dict, **sources_dict2}
              elif isinstance(tok2, sqlparse.sql.Identifier) and tok2.normalized.lower() not in cte_aliases:
                if echo: log('+Table = ' + tok2.normalized.lower())
                sources_dict[tok2.normalized.lower()] = tok.parent

          elif isinstance(tok, sqlparse.sql.Identifier) and tok.normalized.lower() not in cte_aliases:
            if echo: log('+Table = ' + tok.normalized.lower())
            sources_dict[tok.normalized.lower()] = tok.parent


          last_kw_join = False

        if tok.is_keyword and tok.normalized == 'WITH':
          cte_mode = True
          last_kw_from = False
        elif tok.is_keyword and tok.normalized == 'GROUP':
          last_kw_join = False
          last_kw_from = False
        elif tok.is_keyword and tok.normalized == 'WHERE':
          last_kw_join = False
          last_kw_from = False
        elif tok.is_keyword and tok.normalized == 'ORDER':
          last_kw_join = False
          last_kw_from = False
        elif tok.is_keyword and tok.normalized == 'CREATE':
          cte_mode = True
          last_kw_from = False
        elif tok.is_keyword and tok.normalized == 'SELECT':
          cte_mode = False
          last_kw_from = False
        elif tok.is_keyword and tok.normalized == 'FROM':
          last_kw_from = True
        elif tok.is_keyword and 'JOIN' in tok.normalized:
          last_kw_join = True

        last_tok = tok
      done = True
    return sources_dict

  for s, statement in enumerate(statements):
    has_from = False
    last_kw_create = False
    last_kw_create_table = False
    create_table = None

    for tok in statement.tokens:
      if isinstance(tok, sqlparse.sql.Identifier) and last_kw_create_table:
        create_table = tok.normalized
        last_kw_create_table = False
        last_kw_create = False
        if echo: log('-CREATE TABLE ' + create_table)
      if tok.is_keyword and tok.normalized == 'TABLE' and last_kw_create:
        last_kw_create_table = True
      if tok.is_keyword and tok.normalized == 'CREATE':
        last_kw_create = True
      if tok.is_keyword and tok.normalized == 'FROM':
        has_from = True
      last_tok = tok

    if has_from:
      sources_dict = get_sources(statement)
      if create_table:
        sql_sources[create_table] = sorted(sources_dict)
      else:
        sql_sources[s] = sorted(sources_dict)

  return sql_sources



import datetime

from xutil.database.base import DBConn

from xutil.helpers import (get_exception_message, now, log, struct,
                           is_gen_func, isnamedtupleinstance)
from xutil.diskio import read_yaml, write_jsonl, read_jsonl


class OracleConn(DBConn):
  "Oracle Connection"

  data_map = dict(
    string='VARCHAR2()',
    integer='NUMBER()',
    decimal='NUMBER()',
    date='DATE',
    datetime='DATE',
    # timestamp='TIMESTAMP(6)',
    timestamp='DATE',
    text='NCLOB',
  )

  truncate_datef = lambda self, f: 'trunc({})'.format(f)

  def set_variables(self):
    import cx_Oracle
    self.connection.autocommit = False

    self.col_name_id = 2

    self.reverse_data_map = {
      cx_Oracle.NUMBER: 'integer',
      cx_Oracle.STRING: 'string',
      cx_Oracle.LONG_STRING: 'string',
      cx_Oracle.NCHAR: 'string',
      cx_Oracle.FIXED_CHAR: 'string',
      cx_Oracle.DATETIME: 'datetime',
      cx_Oracle.TIMESTAMP: 'datetime',
      cx_Oracle.NATIVE_FLOAT: 'double',
      cx_Oracle.LOB: 'text',
      cx_Oracle.CLOB: 'text',
      cx_Oracle.BLOB: 'text',
      cx_Oracle.NCLOB: 'text',
      cx_Oracle.BINARY: 'text',
    }

    self.error_msg = dict(table_not_exist='table or view does not exist', )

  def _fix_f_name(self, fname):
    "Check length or field, or something else"
    if len(fname) > 30:
      fname = fname[:30]
    return fname

  # def connect(self):
  #   "Connect / Re-Connect to Database"
  #   import cx_Oracle

  #   # get_conn_str = lambda cred: '{}/{}@{}'.format(cred.user, cred.password, cred.name)

  #   def get_conn_str(cred):
  #     if 'service' in cred:
  #       dns_str = cx_Oracle.makedsn(
  #         cred.host, cred.port, service_name=cred.service)
  #     elif 'sid' in cred:
  #       dns_str = cx_Oracle.makedsn(cred.host, cred.port, sid=cred.sid)
  #     return '{}/{}@{}'.format(cred.user, cred.password, dns_str)

  #   conn_str = get_conn_str(struct(self._cred)) if isinstance(
  #     self._cred, dict) else self._cred
  #   self.connection = cx_Oracle.connect(conn_str)
  #   self.cursor = None

  #   self.connection.autocommit = True
  #   self.name = self._cred['name'] if isinstance(
  #     self._cred, dict) else self.connection.tnsentry
  #   self.username = self.connection.username

  def get_dialect(self, echo=False):
    """SQLAlchemy dialect"""
    from sqlalchemy.dialects import oracle, postgresql
    return oracle

  def create_engine(self, conn_str=None, echo=False):
    from cx_Oracle import makedsn
    import sqlalchemy
    if conn_str:
      conn_str = 'oracle+cx_oracle://' + conn_str
    else:
      cred = struct(self._cred)
      if 'service' in cred:
        dns_str = makedsn(cred.host, cred.port, service_name=cred.service)
      elif 'sid' in cred:
        dns_str = makedsn(cred.host, cred.port, sid=cred.sid)
      else:
        dns_str = makedsn(cred.host, cred.port)

      conn_str = (
        'oracle+cx_oracle://{user}:{password}@' + dns_str).format(**cred)

    self.engine = sqlalchemy.create_engine(conn_str, pool_size=10, echo=echo)

    return self.engine

  def insert(self,
             table,
             data,
             field_types=None,
             commit=True,
             fields=[],
             echo=False):
    "Insert records of namedtuple or dicts"
    import cx_Oracle

    s_t = datetime.datetime.now()
    cx_data_map = dict(
      string=cx_Oracle.STRING,
      integer=cx_Oracle.NUMBER,
      decimal=cx_Oracle.NUMBER,
      date=cx_Oracle.DATETIME,
      datetime=cx_Oracle.DATETIME,
      text=cx_Oracle.CLOB,
    )

    # if not len(data):
    #   return False

    row = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(row) else 'dict'
    if not fields:
      fields = row._fields if mode == 'namedtuple' else sorted(row.keys())
    values = [i + 1
              for i in range(len(fields))] if mode == 'namedtuple' else fields

    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()
    
    sql = self._template('core.insert').format(
      table=table,
      options=self._template('core.insert_option'),
      names=', \n'.join([self._fix_f_name(f) for f in fields]),
      values=', \n'.join([':' + str(val) for val in values]),
    )

    if field_types:
      input_sizes = {
        self._fix_f_name(f): cx_data_map[field_types[f][0]]
        for f in field_types
      }
      print(len(input_sizes))
      print(input_sizes)
      cursor.setinputsizes(**input_sizes)

    cursor.prepare(sql)

    try:
      counter = 0
      if is_gen_func(data):
        batch = [row]

        for row in data:
          batch.append(row)
          if len(batch) == self.batch_size:
            cursor.executemany(None, batch)
            counter += len(batch)
            batch = []

        if len(batch):
          cursor.executemany(None, batch)
          counter += len(batch)
      else:
        # cursor.bindvars = None
        cursor.executemany(None, data)
        counter += len(data)

      if commit:
        connection.commit()
      else:
        return counter

    except Exception as e:
      log(Exception('Error for SQL: ' + sql))
      raise e

    finally:
      cursor.close()
      connection.commit()
      connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter, table, mins, rate))
    return counter

  def replace(self,
              table,
              data,
              pk_fields,
              field_types=None,
              commit=True,
              echo=True):
    "Insert/Update records of namedtuple or dicts"
    import cx_Oracle

    s_t = datetime.datetime.now()
    cx_data_map = dict(
      string=cx_Oracle.STRING,
      integer=cx_Oracle.NUMBER,
      decimal=cx_Oracle.NUMBER,
      date=cx_Oracle.DATETIME,
      datetime=cx_Oracle.DATETIME,
      text=cx_Oracle.CLOB,
    )

    if not len(data):
      return False

    row = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(row) else 'dict'
    fields = row._fields if mode == 'namedtuple' else sorted(row.keys())
    values = [i + 1
              for i in range(len(fields))] if mode == 'namedtuple' else fields

    
    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()

    pk_fields_set = set(pk_fields)
    sql = self._template('core.replace').format(
      table=table,
      name_values=',\n'.join(
        [':{} as {}'.format(values[i], f) for i, f in enumerate(fields)]),
      src_tgt_condition='\nAND '.join(
        ['src.{f} = tgt.{f}'.format(f=f) for f in pk_fields]),
      set_fields=',\n'.join([
        'tgt.{f} = src.{f}'.format(f=f) for i, f in enumerate(fields)
        if f not in pk_fields_set
      ]),
      names=',\n'.join(['tgt.{f}'.format(f=f) for f in fields]),
      values=',\n'.join(['src.{f}'.format(f=f) for f in fields]),
    )

    if field_types:
      input_sizes = {
        self._fix_f_name(f): cx_data_map[field_types[f][0]]
        for f in field_types
      }
      print(len(input_sizes))
      print(input_sizes)
      cursor.setinputsizes(**input_sizes)

    cursor.prepare(sql)

    try:
      counter = 0
      if is_gen_func(data):
        batch = [row]

        for row in data:
          batch.append(row)
          if len(batch) == self.batch_size:
            cursor.executemany(None, batch)
            counter += len(batch)
            batch = []

        if len(batch):
          cursor.executemany(None, batch)
          counter += len(batch)
      else:
        # cursor.bindvars = None
        cursor.executemany(None, data)
        counter += len(data)

      if commit:
        connection.commit()
      else:
        return counter

    except Exception as e:
      log(Exception('Error for SQL: ' + sql))
      raise e

    finally:
      cursor.close()
      connection.commit()
      connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter, table, mins, rate))
    return counter

  def _concat_fields(self, fields, as_text=False):
    if as_text:
      fields = ['cast({} as varchar2(4000))'.format(f) for f in fields]
    return ' || '.join(fields)



import csv
import datetime
import os
import re
from collections import namedtuple
from zipimport import zipimporter
import psutil, shutil
import pyarrow as pa
import pyarrow.parquet as pq

import yaml

from yaml import (
  Loader,
  SafeLoader,
)
from ruamel.yaml import YAML

from xutil.helpers import log, now

try:
  from StringIO import StringIO, BytesIO
except:
  from io import StringIO, BytesIO  # python 3

try:
  from itertools import imap
except ImportError:  # Python 3
  imap = map


def construct_yaml_str(self, node):
  # Override the default string handling function
  # to always return unicode objects
  return self.construct_scalar(node)


Loader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
SafeLoader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)


def write_yaml(path, dict_):
  """Save a dict to a YAML file"""
  yaml = YAML()
  with open(path, mode='w', encoding='utf8') as f:
    text = yaml.dump(dict_, f)


def get_zip_path(path):
  for i, segment in enumerate(path.split('/')):
    if '.zip' in segment.lower():
      break
  zip_path = '/'.join(path.split('/')[:i + 1])
  return zip_path


def read_yaml(path):
  """Load from a YAML file, return dict"""
  yaml = YAML()
  if '.zip' in path.lower():
    Z = zipimporter(get_zip_path(path))
    file = BytesIO(Z.get_data(path))
    dict_ = yaml.load(file)
    return dict_

  with open(path) as file:
    dict_ = yaml.load(file)
  return dict_


def read_file(file_path, read_lines=False, mode='r', encoding='utf8'):
  """Read text from file"""
  if '.zip' in file_path.lower():
    Z = zipimporter(get_zip_path(file_path))
    stream = BytesIO(Z.get_data(file_path))
    return stream.read()

  encoding = None if 'b' in mode else encoding
  with open(file_path, mode, encoding=encoding) as stream:
    if read_lines:
      return stream.readlines()
    else:
      return stream.read()


def write_file(file_path,
               text,
               echo=False,
               append=False,
               mode=None,
               encoding='utf8'):
  """Write text to a file"""
  if echo: print('Saving to ' + file_path)
  mode = mode if mode else 'a' if append else 'w'
  f = open(file_path, mode, encoding=encoding)
  f.write(text)
  f.close()


def read_csv(file_path,
             delimiter=',',
             quotechar='"',
             mode='r',
             encoding="utf8"):
  """Read CSV from File"""
  s_t = now()

  with open(file_path, mode, encoding=encoding) as f:
    reader = csv.reader(f, delimiter=delimiter, quotechar=quotechar)
    Data = namedtuple("Data", [f.replace(' ', '_') for f in next(reader)])
    # data_rows = [row for row in reader]
    try:
      i = 0
      data_rows = [row for i, row in enumerate(imap(Data._make, reader))]
    except Exception as e:
      print('ERROR at line ' + str(i + 1))
      raise e

  secs = (now() - s_t).total_seconds()
  rate = round(len(data_rows) / secs, 1)
  log("Imported {} rows from {} [{} r/s].".format(
    len(data_rows), file_path, rate))

  return data_rows


def read_csvD(file_path,
              delimiter=',',
              quotechar='"',
              date_cols=[],
              date_format=None,
              echo=True,
              recarray=False,
              detect_date=True):
  "Use Pandas DataFrame"
  # http://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html
  import pandas
  s_t = now()

  # https://stackoverflow.com/questions/17465045/can-pandas-automatically-recognize-dates
  def date_parser(x):
    dd = x
    try:
      dd = pandas.datetime.strptime(str(x), date_format)
    except ValueError:
      pass
    return dd

  date_parser = date_parser if date_format else None

  df = pandas.read_csv(
    file_path,
    delimiter=delimiter,
    parse_dates=date_cols,
    date_parser=date_parser,
    # quoting=csv.QUOTE_MINIMAL ,
    infer_datetime_format=detect_date,
  )
  
  if recarray:
    df = df.to_records()

  for col in df.columns:
    if not detect_date: continue
    if col in date_cols: continue
    if df[col].dtype == 'object':
      try:
        df[col] = pandas.to_datetime(df[col])
      except ValueError:
        pass

  replace_func = lambda col: re.sub(r'_+', '_', re.sub(r'[\]\[. ]', '_', col))
  df = df.rename(columns={col: replace_func(col) for col in df.columns})

  secs = (now() - s_t).total_seconds()
  rate = round(len(df) / secs, 1)
  if echo:
    log("Imported {} rows from {} [{} r/s].".format(len(df), file_path, rate))

  return df


def read_csvS(file_path,
              delimiter=',',
              headers=None,
              quotechar='"',
              encoding="utf8"):
  with open(file_path, 'r', encoding=encoding) as f:
    reader = csv.reader(f, delimiter=delimiter, quotechar=quotechar)
    Data = namedtuple("Row", next(reader)) if not headers else namedtuple(
      "Row", headers)
    cnt = 0
    for row in imap(Data._make, reader):
      cnt += 1
      yield row

  log("Imported {} rows from {}.".format(cnt, file_path))


def write_csv(file_path,
              headers,
              data,
              footer_text=None,
              append=False,
              log=log,
              echo=True,
              file_obj=None,
              encoding="utf8"):
  "Write to CSV, python3 compatible. 'data' must be list of iterables"
  s_t = now()
  mode = 'a' if append else 'w'
  f = file_obj if file_obj else open(
    file_path, mode, newline='', encoding=encoding)
  w = csv.writer(
    f, delimiter=',', quoting=csv.QUOTE_MINIMAL, lineterminator='\n')
  if not append: w.writerow(headers)
  for row in data:
    w.writerow(row)

  if footer_text:
    f.write(footer_text)

  if not file_obj: f.close()

  secs = (now() - s_t).total_seconds()
  rate = round(len(data) / secs, 1)
  if echo:
    log("Wrote {} rows to {} [{} r/s].".format(len(data), file_path, rate))


def write_csvs(file_path,
               data,
               fields=None,
               lineterminator='\n',
               deli=',',
               log=log,
               secs_d=10,
               gzip=False,
               to_str=False,
               encoding="utf8"):
  "Write to CSV, python3 compatible. 'data' must be list of interables"
  s_t = now()
  l_t = now()
  log_dlt = 100
  counter2 = 0
  f1 = open(file_path, 'w', newline='', encoding=encoding)
  w = csv.writer(
    f1,
    delimiter=deli,
    quoting=csv.QUOTE_MINIMAL,
    lineterminator=lineterminator)

  file_name = file_path.split('/')[-1]
  if fields:
    w.writerow(fields)
  else:
    row = next(data)
    w.writerow(row._fields)  # write headers
    w.writerow(row)  # write 1st row

  counter = 0
  for row in data:
    row = [str(v) for v in row] if to_str else row
    w.writerow(row)
    counter += 1
    counter2 += 1

    # progress message
    if counter2 % log_dlt == 0:
      secs_l = (now() - l_t).total_seconds()
      if secs_l >= secs_d:
        secs = (now() - s_t).total_seconds()
        rate = round(counter2 / secs_l, 1)
        mins = round(secs / 60, 1)
        log("{} min ## Writing to {}: {} rows @ {} r/s.".format(
          mins, file_name, counter, rate))
        l_t = now()
        counter2 = 0

  f1.close()

  if gzip:
    log('Gzipping ' + file_path)
    os.system('gzip -f ' + file_path)
    file_path = file_path + '.gz'

  secs = (now() - s_t).total_seconds()
  rate = round(counter / secs, 1)
  log("Wrote: {} rows to {} [{} r/s].".format(counter, file_path, rate))
  return counter


def write_pq(
    file_path,
    dataf,
    partition_cols=None,
    flavor='spark',
    filesystem=None,
    append=False,
    log=log,
):
  "Write to Parquet, python3 compatible. 'data' must be list of interables"
  s_t = now()

  if not append and os.path.exists(file_path):
    shutil.rmtree(file_path, ignore_errors=True)

  table = pa.Table.from_pandas(dataf, nthreads=psutil.cpu_count())
  counter = table.num_rows
  pq.write_to_dataset(
    table,
    root_path=file_path,
    partition_cols=partition_cols,
    flavor=flavor,
    preserve_index=False,
    filesystem=filesystem,
    use_deprecated_int96_timestamps=True,
    compression='snappy')  # will append. delete folder for overwrite

  secs = (now() - s_t).total_seconds()
  rate = round(counter / secs, 1)
  log("Wrote: {} rows to {} [{} r/s].".format(counter, file_path, rate))
  return counter


def write_pqs(
    file_path,
    dataf_chunks,
    partition_cols=None,
    flavor='spark',
    append=False,
    filesystem=None,
    log=log,
    secs_d=10,
):
  "Stream-Write to Parquet, python3 compatible. 'dataf_chunks' must be list of dataframes"
  s_t = now()
  l_t = now()
  log_dlt = 100
  counter = 0
  counter2 = 0

  if not append and os.path.exists(file_path):
    shutil.rmtree(file_path, ignore_errors=True)

  file_name = file_path.split('/')[-1]
  for dataf in dataf_chunks:
    table = pa.Table.from_pandas(dataf, nthreads=psutil.cpu_count())
    counter += table.num_rows
    counter2 += table.num_rows
    pq.write_to_dataset(
      table,
      root_path=file_path,
      partition_cols=partition_cols,
      flavor=flavor,
      preserve_index=False,
      filesystem=filesystem,
      use_deprecated_int96_timestamps=True,
      compression='snappy')  # will append. delete folder for overwrite

    secs_l = (now() - l_t).total_seconds()
    if secs_l >= secs_d:
      secs = (now() - s_t).total_seconds()
      rate = round(counter2 / secs_l, 1)
      mins = round(secs / 60, 1)
      log("{} min ## Writing to {}: {} rows @ {} r/s.".format(
        mins, file_name, counter, rate))
      l_t = now()
      counter2 = 0

  secs = (now() - s_t).total_seconds()
  rate = round(counter / secs, 1)
  log("Wrote: {} rows to {} [{} r/s].".format(counter, file_path, rate))
  return counter


def write_to_excel(file_path, sheet_order, dict_of_datas):
  'Write to Excel worksheets, dict_of_datas is "WHST_NAME":"DATA_NAMED_TUPLE"'
  from openpyxl import Workbook
  wb = Workbook()
  for sht_name in sheet_order:
    data = dict_of_datas[sht_name]
    sht_name = sht_name if len(sht_name) <= 31 else sht_name[-31:]
    ws = wb.create_sheet(title=sht_name)
    if not data: continue
    log(' >> Adding sheet "{}" with {} rows, {} columns'.format(
      sht_name, len(data), len(data[0])))
    ws.append(data[0]._fields)
    for i, row in enumerate(data):
      ws.append(row)

  wb.remove_sheet(wb.get_sheet_by_name('Sheet'))  # remove 1st default sheet
  wb.save(filename=file_path)
  log("Wrote {} sheets to {}.".format(len(sheet_order), file_path))


def get_hdfs(alias='lake'):
  # https://hdfscli.readthedocs.io/en/latest/api.html
  from hdfs import Config
  client = Config().get_client(alias)
  return client


def write_jsonl(file_path, data, log=log, encoding="utf8"):
  "Write JSONL to File"
  import jsonlines
  s_t = now()

  with open(file_path, 'w', encoding=encoding) as f:
    writer = jsonlines.Writer(f)
    writer.write_all(data)

  counter = len(data)
  secs = (now() - s_t).total_seconds()
  rate = round(counter / secs, 1)
  log("Wrote {} rows to {} [{} r/s].".format(counter, file_path, rate))


def write_jsonls(file_path, data, log=log):
  "Sream Write to JSON Lines. 'data' must be namedtuple. schema is a dict of field to data-type"
  import jsonlines
  s_t = now()
  l_t = now()
  msg_dlt = 10000
  counter = 0
  counter2 = 0

  with open(file_path, 'wb') as f:
    w = jsonlines.Writer(f)
    for row in data:
      w.write(row)
      counter += 1
      counter2 += 1

      # progress message
      if counter2 % msg_dlt == 0:
        secs_l = (now() - l_t).total_seconds()
        if secs_l >= 20:
          secs = (now() - s_t).total_seconds()
          rate = round(counter2 / secs_l, 1)
          mins = round(secs / 60, 1)
          log("{} min ## Writing to JSON: {} rows @ {} r/s.".format(
            mins, counter, rate))
          l_t = now()
          counter2 = 0

  secs = (now() - s_t).total_seconds()
  rate = round(counter / secs, 1)
  log("Wrote {} rows to {} [{} r/s].".format(counter, file_path, rate))


def read_jsonl(file_path, log=log, encoding="utf8"):
  "Read from JSONL File"
  import jsonlines
  s_t = now()

  with open(file_path, 'r', encoding=encoding) as f:
    reader = jsonlines.Reader(f)
    data = list(reader.iter())

  counter = len(data)
  secs = (now() - s_t).total_seconds()
  rate = round(counter / secs, 1)
  log("Read {} rows from {} [{} r/s].".format(counter, file_path, rate))

  return data


def get_path_size(p):
  from functools import partial
  prepend = partial(os.path.join, p)
  return sum([(os.path.getsize(f) if os.path.isfile(f) else get_path_size(f))
              for f in map(prepend, os.listdir(p))])



from xutil.database.base import DBConn
from xutil.helpers import struct


class SQLServerConn(DBConn):
  "SQLServer Connection"

  data_map = dict(
    string='VARCHAR2()',
    integer='NUMBER()',
    decimal='NUMBER()',
    date='DATE',
    datetime='DATE',
    text='CLOB',
  )

  truncate_datef = lambda self, f: 'CONVERT(DATE, {})'.format(f)

  def set_variables(self):
    self.batch_size = 50000

    self.error_msg = dict(table_not_exist='table or view does not exist', )

    self.sql_template = dict(
      drop_table='drop table {} cascade constraints purge',
      insert="INSERT {options} INTO {table} ({names}) VALUES ({values})",
      create_table="create table {table}\n({col_types})",
      create_index="create index {index} on {table} ({cols})",
      insert_option='',
      columns='''select
      owner, table_name, column_name, data_type,
      data_length, data_precision, data_scale
      from all_tab_columns
      where owner = '{owner}'
      and table_name =  '{table}'
      order by column_id''',
      analyze='''select count(*) as tot_cnt, {fields} from {table}''',
      duplicates='''
      with t0 as (
        select
          {fields}, COUNT(1) as c
        from {table}
        group by {fields}
        having count(1) > 1
      )
      select count(1) dup_cnt from t0
      ''',
      replace='''merge into {table} tgt
      USING (SELECT {name_values}
              FROM dual) src
      ON ({src_tgt_condition})
      WHEN MATCHED THEN
        UPDATE SET {set_fields}
      WHEN NOT MATCHED THEN
          insert ({names}) values ({values})
      ''')

  def connect(self):
    "Connect / Re-Connect to Database"
    import pyodbc

    cred = struct(self._cred) if isinstance(self._cred, dict) else None

    self.odbc_string = 'DRIVER={odbc_driver};SERVER={host};DATABASE={database};UID={user};PWD={password}'.format(
      odbc_driver=cred.odbc_driver,
      host=cred.host,
      database=cred.database,
      user=cred.user,
      password=cred.password,
    )
    self.connection = pyodbc.connect(self.odbc_string)

    self.cursor = None

    # self.connection.autocommit = True
    self.name = 'sqlserver'
    self.username = cred.user if cred else ''

    cursor = self.get_cursor()

  def get_dialect(self, echo=False):
    """SQLAlchemy dialect"""
    from sqlalchemy.dialects import mssql
    return mssql

  def create_engine(self, conn_str=None, echo=False):
    import sqlalchemy, urllib

    conn_str = ('mssql+pyodbc:///?odbc_connect={}'.format(
      urllib.parse.quote_plus(self.odbc_string)))

    self.engine = sqlalchemy.create_engine(conn_str, echo=echo)

    return self.engine



import datetime

from xutil.database.base import DBConn
from xutil.helpers import struct, log


class MongoDBConn(DBConn):
  "MongoDB connection"

  def connect(self):
    "Connect / Re-Connect to Database"
    import pymongo
    c = struct(self._cred)

    self.conn = pymongo.MongoClient(
      'mongodb://{host}:{port}/{database}'.format(
        host=c.host,
        port=c.port,
        database=c.database,
      ))

    self._cred['user'] = ''

  def get_coll(self, table):
    db, coll_name = table.split('.') if '.' in table else ('local', table)
    return self.conn[db][coll_name]

  def create_index(self, table, fields):
    "Create index on collection"
    db, coll_name = self.get_coll(table)
    coll_name.create_index('id', unique=True)

  def select_table(self, table, where_dict={}, echo=True, log=log):
    "Select a collection, return list of dicts"
    s_t = datetime.datetime.now()

    collection = self.get_coll(table)
    data = [r for r in collection.find(where_dict)]

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(len(data) / secs, 1)
    if echo:
      log(" >>> Got {} rows in {} secs [{} r/s].".format(
        len(data), secs, rate))

    return data

  def select(self, table, where_dict={}, echo=True, log=log):
    return self.select_table(table, where_dict, echo)

  def insert(self, table, data, echo=True):
    "Inserts list of dics"
    s_t = datetime.datetime.now()

    collection = self.get_coll(table)
    # inserted_ids = collection.insert_many(data).inserted_ids
    collection.insert_many(data)

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(len(data) / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        len(data), table, mins, rate))

    return len(data)

  def replace(self,
              table,
              data,
              pk_filter_dict,
              field_types=None,
              commit=True,
              echo=True):
    "Replace list of dics; pk_filter_dict needs to be condition dict"
    # http://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.bulk_write

    from pymongo import InsertOne, DeleteOne, ReplaceOne
    s_t = datetime.datetime.now()

    collection = self.get_coll(table)

    requests = [
      ReplaceOne(pk_filter_dict, row_dict, upsert=True) for row_dict in data
    ]
    result = collection.bulk_write(requests)

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(len(data) / secs, 1)
    if echo:
      log("Replaced {} records into table '{}' in {} mins [{} r/s].".format(
        len(data), table, mins, rate))

    return len(data)



from distutils.core import setup
from setuptools import find_packages
import os

jdbc = ['jaydebeapi']
hive = [
  "thrift",
  "sasl",  # need sasl.h library
  "thrift_sasl",
  "impyla",
]
web = [
  "python-socketio",
  "eventlet",
  "scrapy",
  "Flask-SSLify"
]
aws = [
  "PyAthenaJDBC"
]
spark = [
  "pyspark"
]

version = '0.1.3'

setup(
  name='xutil',
  # packages=['xutil'],
  version=version,
  description='Extra Useful Utilities.',
  author='Fritz Larco',
  author_email='flarco@live.com',
  url='https://github.com/flarco/xutil',
  download_url='https://github.com/flarco/xutil/archive/master.zip',
  keywords=['xutil'],
  packages=find_packages(exclude=['tests']),
  include_package_data=True,
  test_suite='xutil.tests',
  long_description=open(os.path.join(os.path.dirname(__file__),
                                     'README.rst')).read(),
  install_requires=[
    "verboselogs", "coloredlogs", "psutil", "jmespath", "jsonlines",
    "paramiko", "redis", "cx_Oracle", "psycopg2-binary", "pymongo", "requests",
    "hdfs", "sqlalchemy", "prettytable", "pyarrow", "s3fs",
    "sqlparse", "scp", "ruamel.yaml", "pandas", "recordclass"
  ],
  extras_require={
    'jdbc': jdbc,  # needs gcc and g++ to be installed
    'hive': hive,
    'web': web,
    'full': jdbc + hive + web + aws,
  },
  entry_points={
    'console_scripts': [
      'pykill=xutil.cli:pykill',
      'exec-sql=xutil.cli:exec_sql',
      'exec-etl=xutil.cli:exec_etl',
      'xutil-alias=xutil.cli:alias_cli',
      'xutil-create-profile=xutil.cli:create_profile',
      'ipy=xutil.cli:ipy',
      'xutil-profile-entry=xutil.cli:profile_entry',
      'ipy-spark=xutil.cli:ipy_spark',
    ],
  },
  classifiers=[
    'Programming Language :: Python :: 3', 'Intended Audience :: Developers',
    'Intended Audience :: Education', 'Intended Audience :: Science/Research',
    'Operating System :: MacOS', 'Operating System :: Unix',
    'Topic :: Utilities'
  ])



import sys, datetime
from xutil.database.base import DBConn
from xutil.helpers import get_exception_message, log, is_gen_func, isnamedtupleinstance


class SQLiteConn(DBConn):
  "SQLite Connection"

  data_map = dict(
    string='TEXT',
    integer='NUMBER',
    number='NUMBER',
    decimal='REAL',
    date='TEXT',
    datetime='TEXT',
    text='BLOB',
  )

  def set_variables(self):
    self.error_msg = dict(table_not_exist='table or view does not exist', )

    self.sql_template = dict(
      drop_table='drop table {} cascade constraints',
      insert="INSERT {options} INTO {table} ({names}) VALUES ({values})",
      replace='''REPLACE INTO {table} ({names})
      VALUES({values})''',
      create_table="create table if not exists {table}\n({col_types})",
      create_index="create index {index} on {table} ({cols})",
      insert_option='',
    )

    self._cred.name = 'sqlite3'
    self._cred.user = ''

    self._do_execute('PRAGMA temp_store=MEMORY')
    self._do_execute('PRAGMA journal_mode=MEMORY')

  def get_dialect(self, echo=False):
    """SQLAlchemy dialect"""
    from sqlalchemy.dialects import sqlite
    return sqlite

  def create_engine(self, conn_str=None, echo=False):
    import sqlalchemy
    if not conn_str:
      conn_str = ('sqlite:///' + self._cred.database)

    self.engine = sqlalchemy.create_engine(conn_str, connect_args={'timeout': 7}, echo=echo)

    return self.engine

  def replace(self, table, data, pk_fields=None, commit=True, echo=True):
    "Upsert data into database"
    s_t = datetime.datetime.now()

    if not len(data):
      return False

    row = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(row) else 'dict'
    fields = row._fields if mode == 'namedtuple' else sorted(row.keys())
    values = [i + 1
              for i in range(len(fields))] if mode == 'namedtuple' else fields

    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()

    sql = self._template('core.replace').format(
      table=table,
      names=', '.join(fields),
      values=', '.join(['?' for val in values]),
    )

    try:
      counter = 0
      if is_gen_func(data):
        batch = [row]

        for row in data:
          batch.append(row)
          if len(batch) == self.batch_size:
            cursor.executemany(sql, batch)
            counter += len(batch)
            batch = []

        if len(batch):
          cursor.executemany(sql, batch)
          counter += len(batch)
      else:
        # cursor.bindvars = None
        cursor.executemany(sql, data)
        counter += len(data)

      if commit:
        connection.commit()
      else:
        return counter

    except Exception as e:
      log(Exception('Error for SQL: ' + sql))
      raise e

    finally:
      cursor.close()
      connection.commit()
      connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter, table, mins, rate))
    return counter

  def update(
      self,
      table,
      data,
      pk_fields,
      commit=True,
      echo=True,
      temp_table=None,
  ):
    "Update data in database"
    s_t = datetime.datetime.now()

    if not len(data):
      return False

    row = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(row) else 'dict'
    fields = row._fields if mode == 'namedtuple' else sorted(row.keys())
    if mode == 'namedtuple' and not temp_table:
      data = [r._asdict() for r in data]
      mode = 'dict'

    if temp_table:
      raise Exception('temp_table UPDATE is not supported in SQLite.')

    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()

    pk_fields_set = set(pk_fields)
    sql_tmpl = 'core.update'
    sql_tmpl = sql_tmpl + '_temp' if temp_table else sql_tmpl

    sql = self._template(sql_tmpl).format(
      table=table,
      set_fields=',\n'.join([
        '{f} = :{f}'.format(f=f) for i, f in enumerate(fields)
        if f not in pk_fields_set
      ]),
      pk_fields_equal=' and '.join(
        ['{f} = :{f}'.format(f=f) for f in pk_fields]),
      set_fields2=',\n'.join([
        '{f} = t2.{f}'.format(f=f) for i, f in enumerate(fields)
        if f not in pk_fields_set
      ]),
      pk_fields_equal2=' and '.join(
        ['t1.{f} = t2.{f}'.format(f=f) for f in pk_fields_set]),
      temp_table=temp_table,
    )

    if temp_table:
      # drop / create temp table
      self.drop_table(temp_table)
      self.execute(
        'create table {} as select * from {} where 1=0'.format(
          temp_table,
          table,
        ),
        echo=False,
      )
      self.insert(temp_table, data, echo=False)
      self.execute(sql, echo=False)
      self.execute('drop table if exists ' + temp_table, echo=False)
      counter = len(data)

    else:
      try:
        counter = 0
        if is_gen_func(data):
          batch = [row]

          for row in data:
            batch.append(row)
            if len(batch) == self.batch_size:
              cursor.executemany(sql, batch)
              counter += len(batch)
              batch = []

          if len(batch):
            cursor.executemany(sql, batch)
            counter += len(batch)
        else:
          # cursor.bindvars = None
          cursor.executemany(sql, data)
          counter += len(data)

        if commit:
          connection.commit()
        else:
          return counter

      except Exception as e:
        log(Exception('Error for SQL: ' + sql))
        raise e

      finally:
        cursor.close()
        connection.commit()
        connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter, table, mins, rate))
    return counter

  def insert(self, table, data, echo=False):
    "Insert records of namedtuple or dicts"
    if not data:
      return False

    s_t = datetime.datetime.now()
    row = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(row) else 'dict'
    fields = row._fields if mode == 'namedtuple' else sorted(row.keys())
    values = [i + 1
              for i in range(len(fields))] if mode == 'namedtuple' else fields

    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()

    sql = self._template('core.insert').format(
      table=table,
      options=self._template('core.insert_option'),
      names=', '.join(fields),
      values=', '.join(['?' for val in values]),
    )
    # cursor.prepare(sql)
    try:
      counter = 0
      if is_gen_func(data):
        batch = [row]

        for row in data:
          batch.append(row)
          if len(batch) == self.batch_size:
            cursor.executemany(None, batch)
            counter += len(batch)
            batch = []

        if len(batch):
          cursor.executemany(sql, batch)
          counter += len(batch)
      else:
        cursor.executemany(sql, data)
        counter += len(data)

      connection.commit()
      if echo:
        log("Inserted {} records into table '{}'.".format(counter, table))

    except Exception as e:
      log(Exception('Error for SQL: ' + sql))
      raise e

    finally:
      cursor.close()
      connection.commit()
      connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter,
        table,
        mins,
        rate,
      ))
    return counter



from multiprocessing import Process

import redis, json

from xutil.helpers import jdumps, log


class RedisConn:
  def __init__(self, host, port, db, hkey=None):
    self.conn = redis.StrictRedis(host=host, port=port, db=db)
    self.hkey = hkey
    self.new_pipeline()
    self.pubsub = self.conn.pubsub()
  
  def new_pipeline(self):
    self.pipe = self.conn.pipeline()
    return self.pipe
  
  def set(self, key, val, pipe=False):
    "Set val for key"
    if pipe:
      return self.pipe.set(key, val)
    else:
      return self.conn.set(key, val)
  
  def get(self, key):
    "Get val of key"
    val = self.conn.get(key)
    return val.decode() if val else val
  
  def delete(self, key):
    "Delete key"
    return self.conn.delete(key)
  
  def append(self, key, val, pipe=False):
    "Append val on key"
    if pipe:
      return self.pipe.append(key, val)
    else:
      return self.conn.append(key, val)
  
  def setj(self, key, obj, pipe=False):
    "Set object as JSON string for key"
    if pipe:
      return self.pipe.set(key, jdumps(obj))
    else:
      return self.conn.set(key, jdumps(obj))
  
  def getj(self, key):
    "Get object from JSON string of key"
    val = self.conn.get(key)
    return json.loads(val.decode()) if val else val
  
  def hset(self, key, val, hkey=None, pipe=False):
    "Set val for key of hash"
    hkey = hkey if hkey else self.hkey
    if pipe:
      return self.pipe.hset(hkey, key, val)
    else:
      return self.conn.hset(hkey, key, val)
  
  def hget(self, key, hkey=None):
    "Get val of key of hash"
    hkey = hkey if hkey else self.hkey
    val = self.conn.hget(hkey, key)
    return val.decode() if val else val
  
  def hdel(self, hkey, *names):
    "Delete key in hash"
    hkey = hkey if hkey else self.hkey
    return self.conn.hdel(hkey, *names)
  
  def hgetall(self, hkey):
    "Get all key/val of hash"
    val_dict = self.conn.hgetall(hkey)
    val_dict = {k.decode(): val_dict[k].decode() for k in val_dict} if val_dict else val_dict
    return val_dict
  
  def hsetj(self, key, obj, hkey=None, pipe=False):
    "Set object as JSON string for key of hash"
    hkey = hkey if hkey else self.hkey
    if pipe:
      return self.pipe.hset(hkey, key, jdumps(obj))
    else:
      return self.conn.hset(hkey, key, jdumps(obj))
  
  def hgetj(self, key, hkey=None):
    "Get object from JSON string of key of hash"
    hkey = hkey if hkey else self.hkey
    val = self.conn.hget(hkey, key)
    return json.loads(val.decode()) if val else val
  
  def publish(self, channel, payload):
    "Publish a message to a channel"
    payload = jdumps(payload)
    self.conn.publish(channel, payload)
  
  def subscribe(self, channel):
    "Subscribe to a channel"
    self.pubsub.subscribe(channel)
    return self.pubsub
  
  def rpush(self, name, *values):
    "RPUSH to List"
    return self.conn.rpush(name, *values)
  
  def lpush(self, name, *values):
    "LPUSH to List"
    return self.conn.lpush(name, *values)
  
  def rpop(self, name):
    "RPUSH to List"
    return self.conn.rpop(name)
  
  def lpop(self, name):
    "LPUSH to List"
    return self.conn.lpop(name)
  
  def llen(self, name):
    "LLEN of List: return length of list"
    return self.conn.llen(name)
  
  def ltrim(self, name, start, end):
    "LTRIM of List: Trim the list ``name``, removing all values not within the slice between ``start`` and ``end``"
    return self.conn.ltrim(name, start, end)
  
  def lrange(self, name, start=0, end=-1):
    "LRANGE of List"
    return self.conn.lrange(name, start, end)
  
  def lrangej(self, name, start=0, end=-1):
    "LRANGE of List"
    return [json.loads(obj.decode()) if obj else obj for obj in self.conn.lrange(name, start, end)]
  
  def lset(self, name, index, value):
    "LSET index of List"
    return self.conn.lset(name, index, value)
  
  def lsetj(self, name, index, obj):
    "LSET index of List"
    return self.conn.lset(name, index, jdumps(obj))
  
  def lrem(self, name, count, value):
    "LREM index of List"
    return self.conn.lset(name, count, value)
  
  def lindex(self, name, index):
    "LINDEX return value of index in List"
    return self.conn.lindex(name, index)
  
  def lindexj(self, name, index):
    "LINDEX return value of index in List"
    obj = self.conn.lindex(name, index)
    return json.loads(obj.decode()) if obj else obj
  
  def rpoplpush(self, src, dst):
    "RPOP a value off of the ``src`` list and atomically LPUSH it on to the ``dst`` list.  Returns the value."
    return self.conn.rpoplpush(src, dst)
  
  def listen(self, channel=None):
    "Listen to subscribed channel, yield data"
    if channel: self.pubsub.subscribe(channel)
    
    init_val = False
    try:
      for payload in self.pubsub.listen():
        try:
          payload_data = json.loads(payload['data'].decode('utf-8'))
          yield payload_data
        except AttributeError:
          pass
    except (KeyboardInterrupt, SystemExit):
      return
  
  def start_listener(self, channel, queue, echo=True):
    log('Running Listener on channel: ' + channel, color='green')
    
    def run_redis_listener(redb, channel, queue):
      for payload in redb.listen(channel):
        queue.put(payload)
        if payload == 'exit': return
    
    self.listener_proc = Process(target=run_redis_listener, args=(self, channel, queue))
    self.listener_proc.start()
    return self.listener_proc
  
  def stop_listener(self, channel):
    log('Stopping listener on channel: ' + channel)
    self.conn.publish(channel, jdumps('exit'))



# General Helpers Lib
import datetime, time
import json, atexit
import os, signal
import re, traceback
import sys, socket, psutil
from collections import namedtuple, OrderedDict
from decimal import Decimal
from pathlib import Path

import dateutil
try:
  from colorama import Fore, Back, Style
  color_enabled = True
except ImportError:
  color_enabled = False

host_name = socket.gethostname()

### LAMBDAS ############################

get_rec = lambda row, headers: struct({h.lower():row[i] for i,h in enumerate(headers)})
now = lambda: datetime.datetime.now()
epoch = lambda: int(time.time())
epoch_mil = lambda: int(time.time() * 1000)
now_plus = lambda **kwargs: datetime.datetime.now() + datetime.timedelta(**kwargs)
now_minus = lambda **kwargs: datetime.datetime.now() - datetime.timedelta(**kwargs)
tdelta_seconds = lambda first, second: (first - second).total_seconds()
tdelta_minutes = lambda first, second: (first - second).total_seconds() / 60
tdelta_hours = lambda first, second: (first - second).total_seconds() / 60 / 60
tdelta_days = lambda first, second: (first - second).total_seconds() / 60 / 60 / 24
ndelta_seconds = lambda past: tdelta_seconds(now(), past)  # now delta seconds
ndelta_minutes = lambda past: tdelta_minutes(now(), past)  # now delta minutes
now_str = lambda: datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
now_file_str = lambda: datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
today_str = lambda: datetime.datetime.now().strftime('%Y%m%d')
date_str = lambda: datetime.datetime.now().strftime('%Y-%m-%d')
get_row_fields = lambda r: r._fields if hasattr(r, '_fields') else r.__fields__
jdumps = lambda obj: json.dumps(obj, cls=MyJSONEncoder)
jloads = lambda obj: json.loads(obj)
jtrans = lambda obj: jloads(jdumps(obj))
is_gen_func = lambda x: str(x).startswith('<generator ')
get_profile = lambda create_if_missing=False: load_profile(create_if_missing=create_if_missing)
get_kw = lambda k, deflt, kwargs: kwargs[k] if k in kwargs else deflt

get_script_path = lambda: os.path.dirname(os.path.realpath(sys.argv[0]))
get_dir_path = lambda f=__file__: os.path.dirname(f)
get_home_path = lambda: os.path.expanduser("~")
get_file_name = lambda f=__file__: os.path.splitext(os.path.basename(f))[0]
file_exists = lambda p: Path(p).exists()
kill_pid = lambda pid: os.kill(pid, signal.SIGTERM)
get_pid_path = lambda name=__file__, folder=get_home_path(): '{}/.{}.PID'.format(folder, name)
make_rec = lambda **d: namedtuple('Rec', d.keys())(**d)
set_nice = lambda n: psutil.Process(os.getpid()).nice(n)

elog = lambda text, show_time=True: log(text, color='red', show_time=show_time)
slog = lambda text, show_time=True: log(text, color='green', show_time=show_time)

## LOGGING ############################

try:
  # https://pypi.org/project/coloredlogs/
  # https://coloredlogs.readthedocs.io/en/latest/api.html#changing-the-log-format
  os.environ[
    'COLOREDLOGS_LOG_FORMAT'] = '[%(hostname)s] %(asctime)s %(levelname)s | %(message)s'
  os.environ['COLOREDLOGS_LOG_FORMAT'] = '%(asctime)s -- %(message)s'
  os.environ['COLOREDLOGS_LOG_FORMAT'] = '%(asctime) s%(sep) s%(message)s'

  import coloredlogs, verboselogs
  logger = verboselogs.VerboseLogger(__name__)

  coloredlogs.DEFAULT_FIELD_STYLES = {
    'asctime': {
      'color': 'white'
    },
    'sep': {
      'color': 'magenta'
    },
    'hostname': {
      'color': 'magenta'
    },
    'levelname': {
      'color': 'black',
      'bold': True
    },
    'name': {
      'color': 'blue'
    },
    'programname': {
      'color': 'cyan'
    }
  }

  coloredlogs.DEFAULT_LEVEL_STYLES = {
    'critical': {
      'color': 'red',
      'bold': True
    },
    'debug': {
      'color': 'green'
    },
    'error': {
      'color': 'red'
    },
    'info': {},
    'notice': {
      'color': 'magenta'
    },
    'spam': {
      'color': 'green',
      'faint': True
    },
    'success': {
      'color': 'green',
      'bold': True
    },
    'verbose': {
      'color': 'blue'
    },
    'warning': {
      'color': 'yellow'
    }
  }
  # By default the install() function installs a handler on the root logger,
  # this means that log messages from your code and log messages from the
  # libraries that you use will all show up on the terminal.
  # coloredlogs.install(level='DEBUG')

  # If you don't want to see log messages from libraries, you can pass a
  # specific logger object to the install() function. In this case only log
  # messages originating from that logger will show up on the terminal.
  coloredlogs.install(level='DEBUG', logger=logger)

  def log(text, color='white', level='INFO', **kwargs):
    extra = dict(sep=' -- ')
    level = 'ERROR' if color in ('red', 'r') else level
    level = 'WARNING' if color in ('yellow', 'y') else level
    level = 'SUCCESS' if color in ('green', 'g') else level
    level = 'NOTICE' if color in ('magenta', 'm') else level

    if isinstance(text, Exception):
      level = 'CRITICAL'
      error = text
      # text = 'ERROR: ' + str(error) + '\n' + ''.join(
      #   traceback.format_exception(
      #     etype=type(error), value=error, tb=error.__traceback__))
      text = 'ERROR: ' + ''.join(
        traceback.format_exception(
          etype=type(error),
          value=error,
          tb=error.__traceback__,
        ))

    elif color == 'white':
      text_ = str(text)
      if text_.startswith('~~'):
        level = 'CRITICAL'
        text = text_[2:]
      elif text_.startswith('~'):
        level = 'ERROR'
        text = text_[1:]
      if text_.startswith('-'):
        level = 'WARNING'
        text = text_[1:]
      if text_.startswith('+'):
        level = 'SUCCESS'
        text = text_[1:]
      if text_.startswith('*'):
        level = 'NOTICE'
        text = text_[1:]

    if level == 'CRITICAL':
      logger.critical(text, extra=extra)
    elif level == 'ERROR':
      logger.error(text, extra=extra)
    elif level == 'WARNING':
      logger.warning(text, extra=extra)
    elif level == 'DEBUG':
      logger.debug(text, extra=extra)
    elif level == 'SUCCESS':
      logger.success(text, extra=extra)
    elif level == 'NOTICE':
      logger.notice(text, extra=extra)
    else:
      logger.info(text, extra=extra)

  elog = lambda text, show_time=True: log(text, level='ERROR', show_time=show_time)
  slog = lambda text, show_time=True: log(text, level='SUCCESS', show_time=show_time)

except:

  def log(text, color='white', show_time=True, new_line=True, **kwargs):
    """Print to stdout"""
    if color_enabled:
      time_str = now_str() + Fore.MAGENTA + ' -- ' if show_time else ''
      color_map = dict(
        red=Fore.RED,
        blue=Fore.BLUE,
        green=Fore.GREEN,
        yellow=Fore.YELLOW,
        magenta=Fore.MAGENTA,
        white=Fore.WHITE,
      )
      text = str(text)
      if color == 'white':
        if text.startswith('-'):
          color = 'yellow'
          text = text[1:]
        if text.startswith('+'):
          color = 'green'
          text = text[1:]
        if text.startswith('*'):
          color = 'magenta'
          text = text[1:]
      line = '{}{}{}'.format(
        time_str, color_map[color.lower()] + str(text) + Style.RESET_ALL, '\n'
        if new_line else '')
    else:
      time_str = now_str() + ' -- ' if show_time else ''
      line = '{}{}{}'.format(time_str, text, '\n' if new_line else '')

    sys.stdout.write(line)
    sys.stdout.flush()

  elog = lambda text, show_time=True: log(text, color='red', show_time=show_time)
  slog = lambda text, show_time=True: log(text, color='green', show_time=show_time)


def kill_pid(pid, signum=signal.SIGTERM):
  if os.getpid() == pid:
    # log('-Cannot Kill own PID ({})'.format(pid))
    return
  try:
    os.kill(pid, signum)
  except ProcessLookupError:
    pass


def register_pid(pid_file,
                 pid=None,
                 kill_if_running=False,
                 clean_atexit=True,
                 exit_queue=None):
  "Register PID and verify already running state"
  from .diskio import write_file
  pid = pid if pid else os.getpid()
  cleanup_pid(pid_file, kill_if_running)
  write_file(pid_file, str(pid))
  signal_exit = lambda signum, frame: exit_queue.put(True)
  if clean_atexit:
    if exit_queue:
      # signal.signal(signal.SIGINT, clean_up_func)
      signal.signal(signal.SIGTERM, signal_exit)
    atexit.register(cleanup_pid, pid_file, kill_if_running=True)


def cleanup_pid(pid_file, kill_if_running=False, exit_queue=None):
  "Cleans up old PID file located in home folder"
  from .diskio import read_file
  import psutil

  if os.path.exists(pid_file):
    old_pid = int(read_file(pid_file).rstrip())
    if not kill_if_running and psutil.pid_exists(old_pid):
      raise Exception('PID {} still running! -> {}'.format(old_pid, pid_file))
    kill_pid(old_pid, 9)
    os.remove(pid_file)
    # log('Cleaned Up Old PID {}'.format(old_pid))


def get_exception_message(append_message='', raw=False):
  """Returns the exception message as a formatted string"""
  import linecache
  import traceback

  exc_type, exc_obj, tb = sys.exc_info()
  f = tb.tb_frame
  lineno = tb.tb_lineno
  filename = f.f_code.co_filename
  linecache.checkcache(filename)
  line = linecache.getline(filename, lineno, f.f_globals)
  message = '-' * 65 + '\n' + 'EXCEPTION IN ({}, LINE {} "{}"): {} \n---\n{}'.format(
    filename, lineno, line.strip(), exc_obj,
    traceback.format_exc()) + '\n' + append_message
  if raw: message = str(exc_obj)
  return message


def get_error_str(err):
  err_type = type(err).__name__
  err_msg = str(err)
  return '{}: {}'.format(err_type, err_msg)


time_start = now()


def get_elapsed_time(time_start=time_start):
  time_end = datetime.datetime.now()
  delta_time = divmod(
    (time_end - time_start).days * 86400 + (time_end - time_start).seconds, 60)
  return str(delta_time[0]) + ' mins ' + str(
    delta_time[1]) + ' seconds elapsed'


pprogress_stdt = now()
pprogress_updt = now()
pprogress_i = 1000


def pprogress(i, total, sec_intvl=0.2, text='Complete', show_time=True):
  "Print loop progress"
  global pprogress_updt, pprogress_stdt, pprogress_i
  pp = lambda: sys.stderr.write('\r{:,.1f}% [{} / {}] {}'.format(100.0*(i)/total, i, total, text))
  if i < pprogress_i:
    pprogress_stdt = now()

  if i == total:
    pp()
    print('')
    if show_time: print(get_elapsed_time(pprogress_stdt))
  elif (now() - pprogress_updt).total_seconds() > sec_intvl:
    pp()
    pprogress_updt = now()
  pprogress_i = i


def ptable(headers, rows, format='string', table=None):
  from prettytable import PrettyTable
  if table:
    table.clear_rows()
  else:
    table = PrettyTable(headers)

  for row in rows:
    table.add_row(row)

  return table.get_string()


### Profile ############################

from xutil.diskio import read_file, write_file, read_yaml, write_yaml


def save_profile(data):
  profl_path = os.getenv('PROFILE_YAML', get_home_path() + '/profile.yaml')
  # profl_path = get_home_path() + '/profile.new.yaml'
  write_yaml(profl_path, data)


def load_profile(raw_text=False, create_if_missing=False):
  if not os.getenv('PROFILE_YAML'):
    def_profl_path = get_home_path() + '/profile.yaml'
    templ_path = get_dir_path(__file__) + '/database/templates/profile.yaml'
    if not file_exists(def_profl_path) and create_if_missing:
      write_file(def_profl_path, read_file(templ_path))
    os.environ['PROFILE_YAML'] = def_profl_path
    # raise Exception("Env Var PROFILE_YAML is not set!")

  if raw_text:
    return read_file(os.getenv('PROFILE_YAML'))

  dict_ = read_yaml(os.getenv('PROFILE_YAML'))
  if 'environment' in dict_:
    for key in dict_['environment']:
      os.environ[key] = dict_['environment'][key]

  return dict_


def get_variables(profile=None):
  profile = profile if profile else get_profile()
  if 'variables' in profile:
    return profile['variables']
  else:
    return {}


def get_databases(profile=None):
  profile = profile if profile else get_profile()
  if 'databases' in profile:
    return profile['databases']
  else:
    raise Exception("Key 'databases' not in profile!")


def get_db_profile(db_name):
  dbs = get_databases()
  if db_name in dbs:
    return struct(dbs[db_name])
  else:
    raise Exception("Could not find Database '{}' in profile.".format(db_name))


## CLASSES & DATA OPS ##########################


class struct(dict):
  """ Dict with attributes getter/setter. """

  def __getattr__(self, name):
    return self[name]

  def __setattr__(self, name, value):
    self[name] = value


class State():
  """ State persisted.
  
    >>> from xutil.helpers import State, now
    >>> state = State()  # will create JSON file or load if exists
    >>> state.put(name='Mark', updated=now())
    >>> state.data['name'] = 'Derrick'  # overwrite 'Mark'
  """

  __checkpoint = 5  # auto-save every update count

  def __init__(self, name=None):
    self._name = name if name else get_file_name(__file__)
    self._path = name if '/' in name or '\\' in name else '{}/{}.state'.format(
      get_dir_path(), self._name)

    self.__chkpnt_cnt = 0
    self.__deleted = False

    self.data = {}

    if Path(self._path).exists():
      self.load()
    else:
      self.save()

    # save at program termination
    atexit.register(self.save)

  def put(self, **kwargs):
    "Alter data"
    for k in kwargs:
      self.data[k] = kwargs[k]
      self.__chkpnt_cnt += 1

    if self.__chkpnt_cnt >= self.__checkpoint:
      self.save()
      self.__chkpnt_cnt = 0

  def save(self):
    "Persist to disk"
    if not self.__deleted:
      with open(self._path, 'w') as file:
        json.dump(self.data, file, cls=MyJSONEncoder)

  def load(self):
    "Load from disk"
    if not self.__deleted:
      with open(self._path) as file:
        self.data = json.load(file, object_hook=MyJSONdateHook)
    else:
      print('State has been deleted.')

  def delete(self):
    "Delete from disk"
    os.remove(self._path)
    self.__deleted = True



class MyJSONEncoder(json.JSONEncoder):
  def default(self, o):
    if isinstance(o, datetime.datetime) or isinstance(o, datetime.date):
      # return o.isoformat()
      return str(o)
    elif isinstance(o, bytes):
      return o.decode(errors='ignore')
    elif isinstance(o, Decimal):
      return float(o)
    try:
      val = json.JSONEncoder.default(self, o)
    except:
      val = str(o)
    return val


def MyJSONdateHook(json_dict):
  for (key, value) in json_dict.items():
    try:
      json_dict[key] = datetime.datetime.strptime(value,
                                                  "%Y-%m-%d %H:%M:%S.%f")
    except:
      pass
  return json_dict


def str_format(text, data, enc_char):
  """This takes a text template, and formats the encapsulated occurences with data keys
  Example: if text = 'My name is $$name$$', provided data = 'Bob' and enc_char = '$$'.
    the returned text will be 'My name is Bob'.
  """
  for key in data:
    val = data[key]
    text = text.replace('{}{}{}'.format(enc_char, str(key), enc_char),
                        str(val))
  return text


def str_rmv_indent(text):
  """This removes the indents from a string"""
  # return text
  if not text: return text
  matches = list(re.finditer(r'^ *', text.strip(), re.MULTILINE))
  if not matches: return text
  min_indent = min([len(m.group()) for m in matches])
  regex = r"^ {" + str(min_indent) + r"}"
  new_text = re.sub(regex, '', text, 0, re.MULTILINE)
  return new_text


def split(data, split_size):
  """Yield successive split_size-sized chunks from data (list or dataframe)."""
  if isinstance(data, list):
    for i in range(0, len(data), split_size):
      yield data[i:i + split_size]
  else:
    for i in range(0, len(data), split_size):
      yield data.iloc[i:i + split_size]


def isnamedtupleinstance(x):
  t = type(x)
  b = t.__bases__
  if len(b) != 1 or b[0] != tuple: return False
  f = getattr(t, '_fields', None)
  if not isinstance(f, tuple): return False
  return all(type(n) == str for n in f)


def get_process_data(p):
  mem_info = p.memory_info()
  cpu_times = p.cpu_times()
  rec = p.as_dict()
  # rec = dict(
  #   date_time=now(),
  #   host_name=host_name,
  #   pid=p.pid,
  #   # cmdline = str(p.cmdline()),
  #   cmdline=' '.join(p.cmdline()).replace('\n', ' ').replace('|', ' '),
  #   cpu_percent=_rec['cpu_percent'] or 0,
  #   cpu_times_user=cpu_times.user,
  #   cpu_times_system=cpu_times.system,
  #   cpu_times_children_user=cpu_times.children_user,
  #   cpu_times_children_system=cpu_times.children_system,
  #   create_time=p.create_time(),
  #   # cwd = p.cwd(),
  #   # exe = p.exe(),
  #   # io_counters = p.io_counters(),
  #   memory_info_rss=mem_info.rss,
  #   memory_info_vms=mem_info.vms,
  #   # memory_info_shared=mem_info.shared,
  #   memory_percent=p.memory_percent(),
  #   name=p.name().replace('\n', ' '),
  #   username=p.username().replace('\n', ' '),
  #   status=p.status().replace('\n', ' '),
  #   num_threads=p.num_threads(),
  # )
  return rec


class DictTree:
  def __init__(self, data, header_only=False, name_processor=None):
    self.data = data
    self.name_processor = name_processor
    self.get_keys_path(header_only=header_only)
    self.TRow = namedtuple(
      'TRow',
      'title action field_name field_type depth_level field_path is_null is_list len'
    )

  def get_keys_path(self, header_only=False):
    "Traverses nested dictionary and returns depth level and corresponding JMESPath search string"
    # http://jmespath.org/specification.html#examples
    # https://github.com/jmespath/jmespath.py
    from pyspark.sql import Row
    fields_parent = OrderedDict()
    DP = namedtuple('DictPath', 'level key path vtype field_name')
    name_processor = self.name_processor if self.name_processor else None

    def get_data_type(val):
      if isinstance(val, dict): return 'dict'
      if isinstance(val, list): return 'list'
      try:
        float(val)
        return 'number'
      except:
        try:
          dateutil.parser.parse(val)
          return 'datetime'
        except:
          return 'string'

    def get_field_name(key, parent):
      prefix = ''
      parent = parent.replace('"', '')

      if (parent.endswith('[*]') and len(parent.split('.')) > 1) or \
          (parent.endswith('[*]') and not parent.startswith('[*]')):
        prefix = parent.replace('[*]', '').split('.')[-1] + '_'

      if name_processor:
        prefix = name_processor(prefix)
        key = name_processor(key)

      f_name_new = f_name = prefix + key
      i = 0
      while f_name_new in fields_parent and fields_parent[f_name_new] != parent:
        i += 1
        f_name_new = f_name + str(i)
      fields_parent[f_name_new] = parent
      return f_name_new

    q = lambda s: '"' + s + '"'

    def get_paths(d1, level=1, parent=''):
      kks = set()

      if isinstance(d1, dict):
        if "Header" in d1:
          keys = ["Header"] + [k for k in d1 if k != "Header"]
        else:
          keys = list(d1)
        for key in keys:
          if key.startswith('@'): continue
          vt = get_data_type(d1[key])
          field_name = get_field_name(
            key, parent) if vt not in ('dict', 'list') else None
          path = '{}{}'.format(parent + '.' if parent else '', q(key))
          kks.add(DP(level, key, path, vt, field_name))
          kks = kks.union(get_paths(d1[key], level=level + 1, parent=path))
          if header_only and key == "Header": return kks
      elif isinstance(d1, list):
        key = '[*]'
        path = '{}{}'.format(parent if parent else '', key)
        if level == 1:
          vt = get_data_type(d1)
          kks.add(DP(level, key, path, vt, None))
        for item in d1:
          vt = get_data_type(item)
          kks = kks.union(get_paths(item, level=level + 1, parent=path))

      return kks

    self.paths = sorted(get_paths(self.data))
    self.v_paths = [r for r in self.paths if r.vtype not in ('dict', 'list')]
    self.v_lists = [r for r in self.paths if r.vtype in ('list')]
    self.fields = [
      r.field_name for r in self.paths if r.vtype not in ('dict', 'list')
    ]
    self.fields_dict = {
      r.field_name: r
      for r in self.paths if r.vtype not in ('dict', 'list')
    }
    self.Rec = Row(*self.fields)

    try:
      self.title = list(self.data)[0]
      self.action = self.get_value('Action', conv_date=False)
    except:
      pass

    return self.paths

  def search(self, path):
    from jmespath import search
    return search(path, self.data)

  def get_value(self, field, conv_date=True):
    val = self.search(self.fields_dict[field].path)
    if conv_date and self.fields_dict[field].vtype == 'date':
      try:
        val = dateutil.parser.parse(val)
      except:
        pass
    return val

  def get_field_table(self):
    "Create fields table"
    t_rows = []
    for field in self.fields:
      val = self.get_value(field)
      rec = dict(
        title=self.title,
        action=self.action,
        field_name=field,
        field_type=self.fields_dict[field].vtype,
        depth_level=self.fields_dict[field].level,
        field_path=self.fields_dict[field].path,
        is_null=0 if str(val).strip() else 1,
        is_list=1 if isinstance(val, list) else 0,
        len=len(str(val)),
      )

      t_rows.append(self.TRow(**rec))
    return t_rows

  def get_record(self):
    "Returns one record, omitting lists N:N relationships"
    return self.Rec(*[self.get_value(f) for f in self.fields])



import datetime, os, sys, time, psutil, re, socket
from collections import OrderedDict, namedtuple
import socket
from pathlib import Path

from xutil.database.base import get_conn
from xutil.database.hive import HiveConn, Beeline
from xutil.database.oracle import OracleConn
from xutil.database.postgresql import PostgreSQLConn
from xutil.database.sqlite import SQLiteConn
from xutil.helpers import get_exception_message, now, log, struct, slog, get_profile, get_db_profile, get_kw, get_home_path
from xutil.diskio import read_yaml, write_jsonl, read_jsonl, get_path_size, read_file


class Spark:
  "Spark helper functions"

  version = None

  def __init__(self,
               app_name=None,
               master=None,
               conf={},
               spark_home=None,
               restart=False,
               hive_enabled=False,
               config_name=socket.gethostname().lower(),
               prog_handler=None,
               log=log):

    # restart = True if version != self.version else restart
    if os.getenv('KLOG'): os.system('bash $KLOG')  # kerb login
    spark_home = self.set_sparkenv(spark_home)

    from pyspark import SparkContext, SQLContext, SparkConf
    from pyspark.sql import SparkSession
    active_sc = SparkContext._active_spark_context

    if active_sc:
      log("Active SC ->> " + active_sc.appName)
      sc = active_sc
      spark = SparkSession(sc)
    else:
      sc = None
      spark = None

    if sc and restart:
      log('~Stopping Spark Instance ({})'.format(sc.appName))
      try:
        ps_data = {p.pid: p for p in psutil.process_iter() if p.cmdline()}
        child_pid = ps_data[os.getpid()].children()[0].pid
        if not hive_enabled:
          os.system('kill -9 ' + str(child_pid))
          SparkContext._gateway = None
      except:
        print(get_exception_message())

      sc.stop()
      sc = None

      # sc = sc.getOrCreate()

    profile = get_profile()
    if profile:
      conf_def = profile['spark-conf']
      if 'spark-conf-name' in profile:
        if config_name in profile['spark-conf-name']:
          # overwrite the default spark-conf
          for key in profile['spark-conf-name'][config_name]:
            conf_def[key] = profile['spark-conf-name'][config_name][key]
    else:
      conf_def = {
        "spark.master": "local[4]",
        "spark.driver.memory": "5g",
        "spark.driver.maxResultSize": "2g",
        "spark.driver.cores": "1",
        "spark.executor.instances": "4",
        "spark.executor.cores": "4",
        "spark.sql.broadcastTimeout": 900,
        # "spark.sql.tungsten.enabled": "true",
        "spark.io.compression.codec": "snappy",
        "spark.rdd.compress": "true",
        "spark.streaming.backpressure.enabled": "true",
        "spark.sql.parquet.compression.codec": "snappy",
      }

    # set extraClassPath
    conf_def["spark.driver.extraClassPath"] = self._get_jar_paths(profile)
    if 'SPARK_CLASSPATH' in os.environ and os.environ['SPARK_CLASSPATH']:
      conf_def["spark.driver.extraClassPath"] = conf_def["spark.driver.extraClassPath"] + ':' + os.environ['SPARK_CLASSPATH']
      del os.environ['SPARK_CLASSPATH']

    if master: conf['spark.master'] = master
    if hive_enabled: conf["spark.sql.catalogImplementation"] = "hive"

    for c in conf_def:
      conf[c] = conf_def[c] if c not in conf else conf[c]

    # Launch Spark Instance
    version = self.get_spark_version(spark_home)

    app_name = app_name if app_name else 'Spark_{}_{}_{}'.format(
      str(version).replace('.', ''), os.getenv('USER'), os.getpid())

    if not sc:
      log('Starting Spark Instance ({}) with version {} / {}'.format(
        app_name, version, conf['spark.master']))
      sc, spark, proc = self.init_spark(app_name, spark_home, hive_enabled, conf, restart, prog_handler)
      self.proc = proc

    self.hive_enabled = hive_enabled
    self.version = version
    self.sc = sc
    self.uiWebUrl = sc.uiWebUrl
    self.local_uiWebUrl = 'http://{}:{}'.format(socket.gethostname(), sc.uiWebUrl.split(':')[-1])
    self.spark = spark


  @classmethod
  def set_sparkenv(cls, spark_home=None):
    import urllib.request, subprocess
    URL = 'http://apache.claz.org/spark/spark-2.4.3/spark-2.4.3-bin-hadoop2.7.tgz'
    spark_home = spark_home or os.environ.get('SPARK_HOME', None)

    if not spark_home:
      # download Spark binary
      ans = input(
        'SPARK_HOME is not found. Would you like download and install in your home folder? (Y or N): '
      )
      if ans.lower() == 'y':
        archive_path = get_home_path() + '/' + URL.split('/')[-1]
        if Path(archive_path).exists():
          log('+Archive "{}" already downloaded.'.format(archive_path))
        else:
          log('+Downloading "{}" to "{}"'.format(URL, archive_path))
          urllib.request.urlretrieve(URL, archive_path)
          os.system('cd {} && tar -xf {}'.format(get_home_path(),
                                                 archive_path))
          pyspark_home = archive_path.replace('.tgz', '/python')
          subprocess.call(
            ['pip', 'install', '--target=' + pyspark_home, 'py4j'])
        spark_home = archive_path.replace('.tgz', '')

    os.environ['SPARK_HOME'] = spark_home

    if not Path(spark_home).exists():
      raise Exception('SPARK_HOME="{}" does not exists.'.format(spark_home))

    # add pyspark of SPARK_HOME to path
    sys.path.append(spark_home + '/python')

    log('-Using SPARK_HOME=' + spark_home)

    return spark_home


  def get_master(self):
    return self.spark.sparkContext.master

  def get_spark_version(self, spark_home):
    try:
      line = read_file(spark_home + '/RELEASE', read_lines=True)[0]
      version = line.split()[1]
    except Exception as E:
      log(E)
      log('-Unable to determine Spark Version.')
      version = 'x.x'
    return version

  def get_app_name(self):
    return self.spark.sparkContext.appName

  def init_spark(self,
                 app_name,
                 sparkHome,
                 hive_enabled,
                 conf={},
                 restart=False,
                 prog_handler=None):
    from pyspark import SparkContext, SQLContext, SparkConf
    from pyspark.sql import SparkSession, HiveContext

    sparkconf = SparkConf()
    for key in conf:
      sparkconf.set(key, conf[key])

    if restart:
      SparkContext._gateway = None

    gateway = proc = None
    # gateway, proc = self._launch_gateway(conf=sparkconf)
    # self._launch_gateway_output_handler(proc, prog_handler)
    sc = SparkContext(
      appName=app_name, sparkHome=sparkHome, conf=sparkconf, gateway=gateway)

    if hive_enabled:
      hiveContext = HiveContext(sc)
      spark = SparkSession.builder.appName(app_name).config(
        conf=sparkconf).enableHiveSupport().getOrCreate()
    else:
      spark = SparkSession(sc)


    return sc, spark, proc

  def _get_jar_paths(self, profile):
    from xutil.database.jdbc import get_jar_path
    if 'drivers' not in profile:
      log(Exception('"drivers" key not in profile!'))
      return

    jar_paths = []
    for db_type in profile['drivers']:
      jar_path = get_jar_path(db_type, profile)
      jar_paths.append(jar_path)

    return ':'.join(jar_paths)

  def _set_spark_classpath(self, profile):
    jar_paths = self._get_jar_paths(profile)
    os.environ['SPARK_CLASSPATH'] = '{}:{}'.format(
      jar_paths, os.environ['SPARK_CLASSPATH']
      if 'SPARK_CLASSPATH' in os.environ else '')

  def _launch_gateway_output_handler(self, proc, prog_handler):
    from threading import Thread
    from multiprocessing import Process, Queue, Lock
    import select, fcntl, signal, atexit

    global exit_queue

    exit_queue = Queue()

    def stream_print(proc, exit_queue, prog_handler):

      def spark_log_handler(text, handle=lambda x:x):
        import re
        handle = lambda x: x if not handle else handle
        regex = r"\[Stage (\d+):(=*)>( *)\((\d+) \+ (\d+)\) \/ (\d+)"
        matches = list(re.finditer(regex, text, re.MULTILINE))
        if len(matches) == 0:
          print(text)
          return

        for matchNum, match in enumerate(matches):
          items = []
          for groupNum in range(1, len(match.groups())+1):
            items += [match.group(groupNum)]
          if len(items) == 6:
            items[1] = len(items[1])
            items[2] = len(items[2])
            prog_dict = dict(
              stage=items[0],
              complete=int(items[3]),
              running=int(items[4]),
              total=int(items[5]),
              percent=int(100*int(items[3])/int(items[5])))
            handle(prog_dict)
            log('Spark processing Stage #{}: {}% / {} Running'.format(prog_dict['stage'], prog_dict['percent'], prog_dict['running']))

      def non_block_read(output):
        fd = output.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        try:
          return output.read().decode(errors='replace').rstrip()
        except:
          return ""

      while exit_queue.empty():
        time.sleep(0.05)
        val = non_block_read(proc.stdout)
        if val: spark_log_handler(val, prog_handler)

    # does not exit properly if Thread is not ended
    # https://stackoverflow.com/a/3765160/2295355
    # self.proc_stdout = Thread(target=stream_print, args=(proc, exit_queue, prog_handler))

    self.proc_stdout = Process(
      target=stream_print, args=(proc, exit_queue, prog_handler))

    def signal_handler(*args, **kwargs):
      global exit_queue
      exit_queue.put(1)

    atexit.register(signal_handler)

    self.proc_stdout.start()

  def _launch_gateway(self, conf=None):
    """
    launch jvm gateway
    :param conf: spark configuration passed to spark-submit
    :return:
    """
    import atexit
    import os
    import sys
    import select
    import signal
    import shlex
    import socket
    import platform
    from subprocess import Popen, PIPE, STDOUT

    if sys.version >= '3':
      xrange = range

    from py4j.java_gateway import java_import, JavaGateway, GatewayClient
    from pyspark.find_spark_home import _find_spark_home
    from pyspark.serializers import read_int

    if "PYSPARK_GATEWAY_PORT" in os.environ:
      gateway_port = int(os.environ["PYSPARK_GATEWAY_PORT"])
    else:
      SPARK_HOME = _find_spark_home()
      # Launch the Py4j gateway using Spark's run command so that we pick up the
      # proper classpath and settings from spark-env.sh
      on_windows = platform.system() == "Windows"
      script = "./bin/spark-submit.cmd" if on_windows else "./bin/spark-submit"
      command = [os.path.join(SPARK_HOME, script)]
      if conf:
        for k, v in conf.getAll():
          command += ['--conf', '%s=%s' % (k, v)]
      submit_args = os.environ.get("PYSPARK_SUBMIT_ARGS", "pyspark-shell")
      if os.environ.get("SPARK_TESTING"):
        submit_args = ' '.join(["--conf spark.ui.enabled=false", submit_args])
      command = command + shlex.split(submit_args)

      # Start a socket that will be used by PythonGatewayServer to communicate its port to us
      callback_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      callback_socket.bind(('127.0.0.1', 0))
      callback_socket.listen(1)
      callback_host, callback_port = callback_socket.getsockname()
      env = dict(os.environ)
      env['_PYSPARK_DRIVER_CALLBACK_HOST'] = callback_host
      env['_PYSPARK_DRIVER_CALLBACK_PORT'] = str(callback_port)

      # Launch the Java gateway.
      # We open a pipe to stdin so that the Java gateway can die when the pipe is broken
      if not on_windows:
        # Don't send ctrl-c / SIGINT to the Java gateway:
        def preexec_func():
          signal.signal(signal.SIGINT, signal.SIG_IGN)

        proc = Popen(
          command,
          stdin=PIPE,
          stdout=PIPE,
          stderr=STDOUT,
          preexec_fn=preexec_func,
          env=env)
      else:
        # preexec_fn not supported on Windows
        proc = Popen(command, stdin=PIPE, env=env)

      gateway_port = None
      # We use select() here in order to avoid blocking indefinitely if the subprocess dies
      # before connecting
      while gateway_port is None and proc.poll() is None:
        timeout = 1  # (seconds)
        readable, _, _ = select.select([callback_socket], [], [], timeout)
        if callback_socket in readable:
          gateway_connection = callback_socket.accept()[0]
          # Determine which ephemeral port the server started on:
          gateway_port = read_int(gateway_connection.makefile(mode="rb"))
          gateway_connection.close()
          callback_socket.close()
      if gateway_port is None:
        raise Exception(
          "Java gateway process exited before sending the driver its port number"
        )

      # In Windows, ensure the Java child processes do not linger after Python has exited.
      # In UNIX-based systems, the child process can kill itself on broken pipe (i.e. when
      # the parent process' stdin sends an EOF). In Windows, however, this is not possible
      # because java.lang.Process reads directly from the parent process' stdin, contending
      # with any opportunity to read an EOF from the parent. Note that this is only best
      # effort and will not take effect if the python process is violently terminated.
      if on_windows:
        # In Windows, the child process here is "spark-submit.cmd", not the JVM itself
        # (because the UNIX "exec" command is not available). This means we cannot simply
        # call proc.kill(), which kills only the "spark-submit.cmd" process but not the
        # JVMs. Instead, we use "taskkill" with the tree-kill option "/t" to terminate all
        # child processes in the tree (http://technet.microsoft.com/en-us/library/bb491009.aspx)
        def killChild():
          Popen(["cmd", "/c", "taskkill", "/f", "/t", "/pid", str(proc.pid)])

        atexit.register(killChild)

    # Connect to the gateway
    gateway = JavaGateway(GatewayClient(port=gateway_port), auto_convert=True)

    # Import the classes used by PySpark
    java_import(gateway.jvm, "org.apache.spark.SparkConf")
    java_import(gateway.jvm, "org.apache.spark.api.java.*")
    java_import(gateway.jvm, "org.apache.spark.api.python.*")
    java_import(gateway.jvm, "org.apache.spark.ml.python.*")
    java_import(gateway.jvm, "org.apache.spark.mllib.api.python.*")
    # TODO(davies): move into sql
    java_import(gateway.jvm, "org.apache.spark.sql.*")
    java_import(gateway.jvm, "org.apache.spark.sql.hive.*")
    java_import(gateway.jvm, "scala.Tuple2")

    return gateway, proc

  def get_type_map(self):
    from pyspark.sql.types import (StringType, TimestampType, DoubleType,
                                   BooleanType, IntegerType, LongType,
                                   StructType, StructField, ArrayType, MapType)

    type_map = {
      'int':
      IntegerType(),
      'integer':
      IntegerType(),
      'bigint':
      LongType(),
      'long':
      LongType(),
      'text':
      StringType(),
      'string':
      StringType(),
      'date':
      TimestampType(),
      'datetime':
      TimestampType(),
      'timestamp':
      TimestampType(),
      'float':
      DoubleType(),
      'double':
      DoubleType(),
      'decimal':
      DoubleType(),
      'boolean':
      BooleanType(),
      'array<string>':
      ArrayType(StringType()),
      'array<bigint>':
      ArrayType(LongType()),
      'array<int>':
      ArrayType(IntegerType()),
      'array<date>':
      ArrayType(TimestampType()),
      'array<timestamp>':
      ArrayType(TimestampType()),
      'array<float>':
      ArrayType(DoubleType()),
      'array<double>':
      ArrayType(DoubleType()),
      'array<decimal>':
      ArrayType(DoubleType()),
      'map<string,array<string>>':
      MapType(StringType(), ArrayType(StringType())),
    }

    return type_map

  def df_to_field_types(self, df, n=50000, data_map=None):
    "Get Field:Types from dataframe dtypes"
    from pyspark.sql.types import StringType

    data_map = data_map if data_map else {
      'int64': 'integer',
      'int': 'integer',
      'bigint': 'integer',
      'float64': 'decimal',
      'float': 'decimal',
      'double': 'decimal',
      'decimal': 'decimal',
      'date': 'date',
      'datetime64[ns]': 'datetime',
      'timestamp': 'timestamp',
      'object': 'string',
      'string': 'string',
      'boolean': 'string',
    }

    # df = df.head(n)
    clean_dftype = lambda t: t.split('(')[0] if '(' in t else t
    type_map = {f: data_map[clean_dftype(df.dtypes[i][1])] for i, f in enumerate(df.columns)}
    field_types = OrderedDict()

    df.registerTempTable("df")

    df_max = self.spark.sql('select {} from df'.format(',\n'.join([
      'max(length({f})) as {f}'.format(f=f) for f in df.columns
    ]))).head().asDict()

    for col in df.columns:
      col_ = col.replace(' ', '_')
      if type_map[col] != 'datetime':
        max_len = df_max[col]

      if type_map[col] in ('date', 'datetime', 'timestamp'):
        field_types[col_] = (type_map[col], 0, 0)
      elif type_map[col] in ('integer'):
        max_len = max_len if max_len else 10
        field_types[col_] = (type_map[col], max_len + 1, 0)
      elif type_map[col] in ('string'):
        max_len = max_len if max_len else 100
        if max_len > 2000:
          type_map[col] = 'text'
          max_len = 0
        elif max_len > 1000:
          max_len = 2000
        field_types[col_] = (type_map[col], max_len * 2, 0)
      elif type_map[col] in ('decimal'):
        max_len = max_len if max_len else 5
        field_types[col_] = (type_map[col], max_len + 8, 7)

    return field_types

  def create_schema(self, fields=None, types=None, schema_path=None, log=log):
    "Create Spark schema from lists 'fields' and 'types'"

    from pyspark.sql.types import (
      StructType,
      StructField,
    )

    type_map = self.get_type_map()

    if schema_path is not None:
      schema_data = read_jsonl(schema_path, log=log)
      fields = [r[0] for r in schema_data]
      types = [r[1] for r in schema_data]

    return StructType(
      [StructField(f, type_map[types[i]], True) for i, f in enumerate(fields)])
    # return [[f, type_map[types[i]]] for i, f in enumerate(fields)]

  def create_df(self, data, fields, types=None):
    """
    data = list of tuples or namedtuples
    fields = column names
    type = field types. If None, all will be String.

    df1 = sparko.create_df(
      [('a',2,3), ('b',5,6)],
      ['f1','f2','f3'],
      ['string', 'int', 'int']
    )
    """
    data1 = self.sc.parallelize(data)
    if types == None: types = ['string' for f in fields]

    schema = self.create_schema(fields, types)

    return self.spark.createDataFrame(data1, schema)

  def create_ngram(self, df, n, input_col, output_col='ngrams'):
    "Generate N-Gram -> https://spark.apache.org/docs/2.2.0/ml-features.html#n-gram"
    from pyspark.ml.feature import NGram

    ngram = NGram(n=n, inputCol=input_col, outputCol=output_col)

    ngram_df = ngram.transform(df)
    return ngram_df

  def remove_stop_words(self, df1, input_col, output_col='filtered'):
    "Remove stop words -> https://spark.apache.org/docs/2.2.0/ml-features.html#stopwordsremover"
    from pyspark.ml.feature import StopWordsRemover
    remover = StopWordsRemover(inputCol=input_col, outputCol=output_col)
    df2 = remover.transform(df1)

    return df2

  def tokenize(self, df1, input_col, output_col='words', pattern=None):
    "Tokenize string -> https://spark.apache.org/docs/2.2.0/ml-features.html#tokenizer"
    from pyspark.ml.feature import Tokenizer, RegexTokenizer

    if pattern:
      tokenizer = RegexTokenizer(
        inputCol=input_col, outputCol=output_col, pattern=pattern)
    else:
      tokenizer = Tokenizer(inputCol=input_col, outputCol=output_col)

    tokenized_df = tokenizer.transform(df1)

    return tokenized_df

  @staticmethod
  def hdfs_put(file_path,
               hdfs_folder,
               put=True,
               rm=True,
               log=log):
    "Move file from local to HDFS"
    file_size = get_path_size(file_path)
    file_name = file_path.split('/')[-1]
    hdfs_file_path = '{}/{}'.format(hdfs_folder, file_name)
    s_t = now()
    if rm:
      os.system('hdfs dfs -rm -r -f -skipTrash {}'.format(hdfs_file_path))
    if put:
      os.system('hdfs dfs -put -f {} {}'.format(file_path, hdfs_folder))
      secs = (now() - s_t).total_seconds()
      mins = round(secs / 60, 1)
      rate = round(file_size / 1024**2 / secs, 1)
      log("Moved data into HDFS in {} mins [{}MB -> {} MB/s]".format(mins, file_size, rate))
    return hdfs_file_path

  def read_json(self,
                hdfs_file_path,
                schema,
                timestampFormat='yyyy-MM-dd HH:mm:ss',
                dateFormat='yyyy-MM-dd',
                date_cols=[],
                timestamp_cols=[],
                log=log):
    "Read from json file on HDFS"

    df = self.spark.read.json(
      hdfs_file_path,
      # timestampFormat=timestampFormat,
      # dateFormat=dateFormat,
      schema=schema,
      # multiLine=multiLine,
      mode='FAILFAST',
    )

    df = self.process_df_fields(df, date_cols, timestamp_cols, dateFormat,
                                timestampFormat)

    return df

  def read_json2(self,
                 file_path,
                 hdfs_folder,
                 schema_path=None,
                 timestampFormat='yyyy-MM-dd HH:mm:ss',
                 dateFormat='yyyy-MM-dd',
                 date_cols=[],
                 timestamp_cols=[],
                 hdfs_put=True,
                 log=log):
    "Read from json file initially on local"

    schema_path = schema_path if schema_path else file_path + '.schema'
    schema_data = read_jsonl(schema_path, log=log)
    schema = self.create_schema([r[0] for r in schema_data],
                                [r[1] for r in schema_data])

    hdfs_file_path = self.hdfs_put(file_path, hdfs_folder, put=hdfs_put)

    return self.read_json(
      hdfs_file_path,
      schema=schema,
      timestampFormat=timestampFormat,
      dateFormat=dateFormat,
      date_cols=date_cols,
      timestamp_cols=timestamp_cols,
      log=log,
    )

  def read_csv2(self,
                file_path=None,
                hdfs_file_path=None,
                hdfs_folder=None,
                timestampFormat='yyyy-MM-dd HH:mm:ss',
                dateFormat='yyyy-MM-dd',
                delimeter=',',
                date_cols=[],
                timestamp_cols=[],
                log=log,
                escape='"',
                hdfs_put=True,
                ignoreTrailingWhiteSpace=False,
                schema=None,
                hdfs_delete_local=False):
    "Read from csv file initially on local"
    # https://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html
    # The dateFormat field does not parse/assign dates correctly... not sure why.
    # see https://stackoverflow.com/questions/40878243/reading-csv-into-a-spark-dataframe-with-timestamp-and-date-types
    # seems the only way is to assign the schema manually or to cast manually (which is being done here).

    if self.hive_enabled and file_path:
      if not hdfs_folder:
        raise('No value provided for hdfs_folder!')

      hdfs_file_path = hdfs_file_path if hdfs_file_path else self.hdfs_put(
        file_path, hdfs_folder, put=self.hive_enabled, log=log)
      if hdfs_delete_local:
        os.system('rm -rf ' + file_path)
      f_path = hdfs_file_path
    else:
      f_path = file_path

    params = dict(
      path=f_path,
      header=True,
      timestampFormat=timestampFormat,
      # dateFormat=dateFormat,
      inferSchema=False if schema else True,
      schema=schema,
      sep=delimeter,
      escape=escape,
      ignoreTrailingWhiteSpace=ignoreTrailingWhiteSpace,
    )

    if self.version >= '2.2': params['multiLine'] = True
    df = self.spark.read.csv(**params)

    df = self.process_df_fields(df, date_cols, timestamp_cols, dateFormat,
                                timestampFormat)

    log('Finished reading file: ' + f_path)
    return df

  def process_df_fields(self, df, date_cols, timestamp_cols, dateFormat,
                        timestampFormat):
    "Process dataframe fields names and date/time conversions"

    for col in df.columns:
      df = df.withColumnRenamed(col, col.lower().replace(' ', '_'))

    if date_cols:
      date_cols = [c.lower() for c in date_cols]
      df.registerTempTable("df")
      date_str = "to_date(cast(unix_timestamp({c}, '{dateF}') as timestamp)) as {c}"
      sql = '''select {} from df'''.format(', '.join([
        date_str.format(c=col, dateF=dateFormat) if col in date_cols else col
        for col in df.columns
      ]))
      df = self.spark.sql(sql)

    if timestamp_cols:
      timestamp_cols = [c.lower() for c in timestamp_cols]
      df.registerTempTable("df")
      date_str = "cast(unix_timestamp({c}, '{dateF}') as timestamp) as {c}"
      sql = '''select {} from df'''.format(', '.join([
        date_str.format(c=col, dateF=timestampFormat)
        if col in timestamp_cols else col for col in df.columns
      ]))
      df = self.spark.sql(sql)

    return df

  @staticmethod
  def write_csv(df,
                file_path,
                timestampFormat='yyyy-MM-dd HH:mm:ss',
                delimeter=',',
                partitions=1,
                log=log):
    "Write to csv file locally"
    df.repartition(partitions).write.csv(
      file_path + '_',
      mode='overwrite',
      header=True,
      timestampFormat=timestampFormat,
      sep=delimeter,
    )
    os.system(
      'cat {file_path}_/*.csv > {file_path} && rm -rf {file_path}_'.format(
        file_path=file_path))

  def write_csv2(self,
                 df,
                 file_path,
                 hdfs_folder = None,
                 timestampFormat='yyyy-MM-dd HH:mm:ss',
                 dateFormat='yyyy-MM-dd',
                 delimeter=',',
                 partitions=1,
                 log=log,
                 quoteAll=False,
                 gzip=False):
    "Write to csv file correctly from HDFS to local"
    file_name = file_path.split('/')[-1]
    if self.hive_enabled:
      f_path = '{}/{}'.format(hdfs_folder, file_name)
    else:
      f_path = file_path
    df.repartition(partitions).write.csv(
      f_path,
      mode='overwrite',
      header=True,
      timestampFormat=timestampFormat,
      dateFormat=dateFormat,
      sep=delimeter,
      quoteAll=quoteAll)
    if self.hive_enabled:
      os.system('hdfs dfs -getmerge {} {}'.format(f_path, file_path))
      os.system('rm -f .{}.crc'.format(file_path))
    elif partitions > 1:
      # merge into one partition
      os.system('cat {file_path}/*.csv > {file_path}_ && rm -rf {file_path} && mv {file_path}_ {file_path}'.format(file_path=file_path))

    if gzip:
      log('Gzipping ' + file_path)
      os.system('gzip -f ' + file_path)
      file_path = file_path + '.gz'

    log("Wrote data to file {}.".format(file_path))

  @staticmethod
  def write_jsonl(df,
                  file_path,
                  hdfs_folder,
                  mode='overwrite',
                  timestampFormat='yyyy-MM-dd HH:mm:ss',
                  dateFormat='yyyy-MM-dd',
                  partitions=1,
                  log=log):
    "Write to csv file correctly from HDFS to local"

    file_name = file_path.split('/')[-1]

    # Write schema
    schema_file_path = file_path + '.schema'
    schema_data = df.dtypes
    write_jsonl(schema_file_path, schema_data)

    # Write data
    df.repartition(partitions).write.json(
      path='{}/{}'.format(hdfs_folder, file_name),
      mode=mode,
      timestampFormat=timestampFormat,
      dateFormat=dateFormat,
    )
    os.system('hdfs dfs -getmerge {}/{} {}'.format(hdfs_folder, file_name,
                                                   file_path))
    crc_file_path = '{}/.{}.crc'.format('/'.join(file_path.split('/')[:-1]),
                                        file_name)
    os.system('rm -f {}'.format(crc_file_path))

    log("Wrote data to file {}.".format(file_path))

  def convert_fields(self,
                     df,
                     curr_type,
                     new_type,
                     conv_str="cast({col} as {new_type}) as {col}"):
    """Convert fields to specified types"""

    df.registerTempTable("df")
    sql = '''select {} from df'''.format(', '.join([
      conv_str.format(col=col[0], new_type=new_type)
      if col[1] == curr_type else col[0] for col in df.dtypes
    ]))
    df = self.spark.sql(sql)

    return df

  def jdbc_read(self, db_name, sql_table, partition_dict={}, fetchsize=10000, log=log):
    "Read data from table or sub-query SQL"
    db_profile = get_db_profile(db_name)
    null_sql_table = None

    if partition_dict:
      # This attempts the split a range evenly
      # This is not optimal when there is no certainty that the range is
      # continuous, such as an incremented BIGINT primary key.

      if not ('partitionColumn' in partition_dict and 'numPartitions' in partition_dict):
        raise("partition_dict must contain 'partitionColumn' and 'numPartitions'")

      # Automatically obtain lowerBound and upperBound values.
      if not ('lowerBound' in partition_dict and 'upperBound' in partition_dict):
        conn = get_conn(db_name, echo=False)
        sql = conn.template('routine.number_min_max').format(
          field=partition_dict['partitionColumn'],
          table=sql_table,
        )
        log('Getting partitionColumn stats for {}'.format(
          partition_dict['partitionColumn']))
        data = conn.select(sql, echo=False)

        if not len(data):
          raise('Table "{}" seems empty?'.format(sql_table))

        tot_cnt, field_cnt, min_val, max_val = data[0]
        null_cnt = tot_cnt - field_cnt

        # if there are null count, they will be missed
        if null_cnt > 0:
          log('null_cnt ({}) is > 0 for field "{}" in table "{}". Selecting a Null filtered partition!'.format(null_cnt, partition_dict['partitionColumn'], sql_table))

          null_sql_table = "(select * from {} where {} is null)".format(sql_table, partition_dict['partitionColumn'])

          df_null = self.spark.read \
            .format("jdbc") \
            .option("url", db_profile['url']) \
            .option("dbtable", null_sql_table) \
            .option("user", db_profile['user']) \
            .option("password", db_profile['password']) \
            .option("fetchsize", fetchsize) \
            .load()

        partition_dict['lowerBound'] = min_val
        partition_dict['upperBound'] = max_val

      df = self.spark.read \
        .format("jdbc") \
        .option("url", db_profile['url']) \
        .option("dbtable", sql_table) \
        .option("user", db_profile['user']) \
        .option("password", db_profile['password']) \
        .option("partitionColumn", partition_dict['partitionColumn']) \
        .option("lowerBound", partition_dict['lowerBound']) \
        .option("upperBound", partition_dict['upperBound']) \
        .option("numPartitions", partition_dict['numPartitions']) \
        .option("fetchsize", fetchsize) \
        .load()
    else:
      df = self.spark.read \
        .format("jdbc") \
        .option("url", db_profile['url']) \
        .option("dbtable", sql_table) \
        .option("user", db_profile['user']) \
        .option("password", db_profile['password']) \
        .option("fetchsize", fetchsize) \
        .load()
      tot_cnt = df.count()

    if null_sql_table:
      df.registerTempTable("df")
      df_null.registerTempTable("df_null")
      df = self.spark.sql('select * from df union all select * from df_null')

    self._last_df_cnt = tot_cnt
    return df

  def sql(self, *args, **kwargs):
    "Run Spark SQL"
    return self.spark.sql(*args, **kwargs)

  def jdbc_write(self,
                 df,
                 db_profile,
                 sql_table,
                 create_table=True,
                 truncate=False,
                 order_by=[],
                 grant_sql=None,
                 post_sql=None,
                 batch_size=20000,
                 partitions=7,
                 tgt_mode='append',
                 log=log,
                 **kwargs):
    "Write data to table"

    count_recs = get_kw('count_recs', False, kwargs)
    counter = df.count() if count_recs else get_kw('tot_cnt', -1, kwargs)


    if db_profile['type'] == 'oracle':
      # trim field names if is Oracle
      for col in df.columns:
        df = df.withColumnRenamed(col, col[:30]) if len(col) > 30 else df
      for col in df.columns:
        df = df.withColumnRenamed(col, col.upper().replace(' ', '_'))

      # convert booleans to string
      df = self.convert_fields(df, 'boolean', 'string')

    s_t = datetime.datetime.now()
    conn = get_conn(db_profile['name'])

    if truncate and tgt_mode != 'overwrite':
      create_table = False
      conn.execute('TRUNCATE TABLE ' + sql_table)
      time.sleep(2)

    if tgt_mode == 'overwrite':
      field_types = self.df_to_field_types(df)
      conn.create_table(sql_table, field_types, drop=True)
      if grant_sql:
        conn.execute(grant_sql)
      time.sleep(2)
      tgt_mode = 'append'

    partitions = partitions if order_by else 1
    df.repartition(partitions).write \
      .format("jdbc") \
      .option("url", db_profile['url']) \
      .option("dbtable", sql_table) \
      .option("user", db_profile['user']) \
      .option("password", db_profile['password']) \
      .option("batchsize", batch_size) \
      .save(mode=tgt_mode)

    # .option("createTableColumnTypes", create_types_str) \ # 2.2 compatible
    # .option("sessionInitStatement", sessionInitStatement) \ # 2.3 compatible
    # .option("customSchema", customSchema) \ # 2.3 compatible

    # if order_by:
    #   # using many partitions, the data is not ordered
    #   sql = '''
    #   CREATE TABLE {t2} AS SELECT * FROM {t1} ORDER BY {ord};
    #   DROP TABLE {t1} CASCADE CONSTRAINTS PURGE;
    #   ALTER TABLE {t2} RENAME TO {t1n};
    #   '''.format(
    #     t1=sql_table,
    #     # t1n=sql_table.split('.')[-1] if '.' in sql_table else sql_table,
    #     t1n=sql_table.split('.')[-1] if '.' in sql_table else sql_table,
    #     t2=sql_table + 'z',
    #     ord=', '.join(order_by),
    #   )
    #   conn.execute(sql, 'ORDERING', echo=False)

    if post_sql:
      conn.execute(post_sql, 'EXECUTING POST-SQL', echo=False)

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)

    log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
      counter, sql_table, mins, rate))

  def hive_write(self,
                 df,
                 table,
                 mode='overwrite',
                 format='parquet',
                 convert_dec=True,
                 max_partitions=200,
                 show_count=True,
                 tot_cnt=None,
                 order_by=[],
                 log=log):
    "Write to Hive"
    s_t = datetime.datetime.now()

    # Parquet is the default format, which does not play nice with Decimal fields
    # best to convert those to double.
    if convert_dec: df = self.convert_fields(df, 'decimal', 'double')

    log("Loading into hive table {}...".format(table))
    if max_partitions and df.rdd.getNumPartitions() > max_partitions:
      df = df.repartition(max_partitions)

    if format == 'orc':
      df.write.format("orc").saveAsTable(table, mode=mode)
    elif format == 'parquet':
      df.write.saveAsTable(table, mode=mode)

    if format == 'parquet' and order_by:
      log('Ordering table ' + table)

      hive_db = read_yaml(os.getenv('PROFILE_YAML'))['databases']['HIVE']
      h_conn = Beeline(hive_db, echo=False)
      h_conn.convert_to_parquet(table, order_by=order_by)

    secs = (datetime.datetime.now() - s_t).total_seconds()

    mins = round(secs / 60, 1)
    if show_count:
      df.registerTempTable("df")
      counter = self.spark.sql('select count(1) cnt from df').head(1)[0][0]
      if tot_cnt and tot_cnt != counter:
        log('-WARNING: table vs. file count differ! {} != {}'.format(
          counter, tot_cnt))
      rate = round(counter / secs, 1)
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter, table, mins, rate))
    else:
      log("Inserted records into table '{}' in {} mins.".format(table, mins))

  def hive_load_large_file(self,
                           file_path,
                           target_table,
                           split_size_GB=5,
                           log=log):
    "Load large files into hive"

    def execute_cmd(cmd):
      rt = os.system(cmd)
      if rt > 0:
        print('ERROR: Return code not 0 !')
        sys.exit(rt)

    file_size = os.path.getsize(file_path)
    file_lines = int(
      os.popen('wc -l "{}"'.format(file_path)).readline().split()[0])
    avg_size_per_line = file_size / file_lines
    lines_split = int(split_size_GB * 1024**3 / avg_size_per_line)

    log('File size is {} MB or {} GB'.format(
      round(1.0 * file_size / (1024**2), 1),
      round(1.0 * file_size / (1024**3), 1),
    ))
    log('Splitting large file...')
    out_folder = file_path + '.split'
    os.system('mkdir -p ' + out_folder)

    # split large file and add headers
    cmd = 'cd {out_folder} && tail -n +2 "{file_path}" | split -l {lines_split} - {output_prefix}'.format(
      out_folder=out_folder,
      file_path=file_path,
      lines_split=lines_split,
      output_prefix=file_path.split('/')[-1] + '.',
    )
    execute_cmd(cmd)

    log('Adding headers...')
    cmd = '''
    cd {out_folder} && for file in *
    do
        head -n 1 {file_path} > tmp_file
        cat $file >> tmp_file
        mv -f tmp_file $file
    done
    '''.format(
      out_folder=out_folder,
      file_path=file_path,
    )
    execute_cmd(cmd)

    files = [out_folder + '/' + fn for fn in os.listdir(out_folder)]

    for i, file_path in enumerate(files):
      log('Loading ' + file_path)
      df1 = self.read_csv2(
        file_path,
        delimeter=',',
        timestampFormat='yyyy-MM-dd HH:mm:ss',
        dateFormat='yyyy-MM-dd',
        date_cols=[])
      self.hive_write(
        df1, target_table, mode='append' if i > 0 else 'overwrite')


class SparkConn(HiveConn):
  "Spark Hive Context Connection"

  df_id = None
  df_ids = []

  def connect(self):
    "Connect / Re-Connect to Database"
    c = struct(self._cred)
    restart = c.restart if 'restart' in c else False
    hive_enabled = c.hive_enabled if 'hive_enabled' in c else False
    master = c.master if 'master' in c else None
    version = c.version if 'version' in c else None
    spark_home = c.spark_home if 'spark_home' in c else None
    self.sparko = Spark(
      restart=restart,
      hive_enabled=hive_enabled,
      master=master,
      spark_home=spark_home)
    self.application_id = self.sparko.sc._jsc.sc().applicationId()

    self._cred.name = self.name = "Spark"
    self.username = c.user

  def get_cursor(self):
    return self.sparko

  def close(self):
    self.sparko.sc.stop()

  def get_cursor_fields(self):
    "Get fields of active Select cursor"

    return self._fields

  def select(self,
             sql,
             rec_name='Record',
             dtype='namedtuple',
             limit=None,
             echo=True,
             log=log):
    "Select from SQL, return list of namedtuples"
    s_t = datetime.datetime.now()

    df1 = self.sparko.sql(sql)
    self.df_id = self.df_id if self.df_id else 'df_' + str(int(time.time()))
    df1.registerTempTable(self.df_id)
    self.df_ids.append(self.df_id)
    df1 = df1.limit(limit) if limit else df1

    if dtype == 'dataframe':
      data = df1.toPandas()
    else:
      data = df1.collect()

    self._fields = df1.columns
    secs = (datetime.datetime.now() - s_t).total_seconds()
    rate = round(len(data) / secs, 1)

    if echo:
      log(" >>> Got {} rows in {} secs [{} r/s].".format(
        len(data), secs, rate))

    return data



# Init
from xutil.helpers import (log, slog, elog)
from xutil.parallelism import (kill_processes)
from xutil.database.base import (get_conn)
from xutil.database.etl import (exec_sql, db_to_db, db_to_ff, ff_to_db)
from xutil.web import (WebApp)